#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Exoplanet archive tools.

@author: Steven Charles-Mindoza
"""

from rich import print
from transitfit import calculate_logg
from datetime import datetime, timedelta
from tabulate import tabulate
from pandas import DataFrame, read_csv, Categorical, concat
from astroquery.mast import Observations as obs
from fuzzywuzzy import process
from natsort import natsorted
from tqdm import tqdm
import numpy as np
import time
import sys
import os
import threading

tqdm.pandas(desc="Progress")


class NaNError(Exception):
    pass

class suppress_print():
    def __enter__(self):
        self.original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self.original_stdout


def _load_csv():
    '''
    Coverts all 4 archives to have the same column structure.

    '''
    download_archive_threaded_execute()
    nasa_csv = 'firefly/data/nasa.csv.gz'
    eu_csv = 'firefly/data/eu.csv.gz'
    oec_csv = 'firefly/data/oec.csv.gz'
    org_csv = 'firefly/data/org.csv.gz'
    here = os.path.dirname(os.path.abspath(__file__))
    spearnet_csv = f'{here}/data/spear.csv'
    spearnet_csv_uncoup = f'{here}/data/spear_uncoupled.csv'
    global exo_nasa, exo_eu, exo_oec, exo_org, exo_spearnet, exo_spearnet_uncoup
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # NASA
    exo_nasa = read_csv(nasa_csv)
    cols = exo_nasa.columns.tolist()
    blank = DataFrame(columns=cols)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # SPEARNET
    exo_spearnet = read_csv(spearnet_csv)
    exo_spearnet = exo_spearnet.append(blank)
    exo_spearnet['archive'] = 'SPEARNET COUPLED'

    exo_spearnet_uncoup = read_csv(spearnet_csv_uncoup)
    exo_spearnet_uncoup = exo_spearnet_uncoup.append(blank)
    exo_spearnet_uncoup['archive'] = 'SPEARNET UNCOUPLED'
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # EU
    col_subset_eu = ['# name', 'orbital_period', 'semi_major_axis', 'radius',
                     'eccentricity', 'inclination', 'tzero_tr', 'star_teff',
                     'star_teff_error_max', 'star_radius',
                     'star_radius_error_max', 'star_mass',
                     'star_metallicity', 'star_metallicity_error_max', 'omega']
    exo_eu = read_csv(eu_csv, usecols=col_subset_eu)
    exo_eu.columns = ['pl_name', 'pl_radj', 'pl_orbper', 'pl_orbsmax',
                      'pl_orbeccen', 'pl_orbincl', 'pl_orblper', 'pl_tranmid', 'st_met',
                      'st_meterr1', 'st_mass','st_rad',
                      'st_raderr1', 'st_teff', 'st_tefferr1']
    exo_eu = exo_eu.append(blank)
    exo_eu['archive'] = 'EU'
    exo_nasa['archive'] = 'NASA'
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # OEC
    col_subset_oec = ['name', 'radius', 'period',
                     'semimajoraxis', 'eccentricity', 'periastron', 'inclination',
                     'hoststar_mass', 'hoststar_radius',
                     'hoststar_metallicity', 'hoststar_temperature']
    exo_oec = read_csv(oec_csv, usecols=col_subset_oec)
    exo_oec.columns = ['pl_name', 'pl_radj', 'pl_orbper', 'pl_orbsmax',
                      'pl_orbeccen', 'pl_orblper','pl_orbincl', 'st_mass',
                      'st_rad', 'st_met', 'st_teff']
    exo_oec = exo_oec.append(blank)
    exo_oec['archive'] = 'OEC'
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # ORG
    col_subset_org = ['NAME', 'R', 'PER',
                     'SEP', 'ECC', 'I', 'TT',
                     'MSTAR', 'RSTAR', 'RSTARUPPER',
                     'FE', 'FEUPPER', 'LOGG', 'LOGGUPPER',
                     'TEFF', 'TEFFUPPER']
    exo_org = read_csv(org_csv, usecols=col_subset_org)
    exo_org.columns = ['pl_orbeccen', 'st_met', 'st_meterr1', 'pl_orbincl',
                      'st_logg', 'st_loggerr1','st_mass', 'pl_name',
                      'pl_orbper', 'pl_radj', 'st_rad', 'st_raderr1', 'pl_orbsmax',
                      'st_teff', 'st_tefferr1', 'pl_tranmid']
    exo_org = exo_org.append(blank)
    exo_org['archive'] = 'ORG'
    

def _search(exoplanet):
    exo_list = exo_nasa[['pl_name', 'tic_id']] \
              .dropna() \
              .drop(['tic_id'], axis=1) .values .tolist()
    exo_list = [j for i in exo_list for j in i]
    exo_list = natsorted(exo_list)
    
    ratios = process.extract(exoplanet,exo_list)
    highest = process.extractOne(exoplanet,exo_list)
    return highest, ratios


def _search_all(exoplanet):
    archive_list = [exo_nasa, exo_eu, exo_oec, exo_org]
    exo_archive = concat(archive_list)
    exo_list = exo_archive[['pl_name']] \
              .dropna() .drop_duplicates('pl_name') \
              .values .tolist()
    exo_list = [j for i in exo_list for j in i]
    exo_list = natsorted(exo_list)
    
    ratios = process.extract(exoplanet,exo_list)
    highest = process.extractOne(exoplanet,exo_list)
    return highest, ratios


def _tic(exoplanet):
    '''
    Returns a dataframe of all planet names and tic ID's.

    Parameters
    ----------
    None

    Returns
    -------
    TIC ID of an exoplanet target.

    '''
    df = exo_nasa[['pl_name','tic_id']] \
        .sort_values('pl_name') \
        .drop_duplicates('pl_name', keep='last')
    tic_df = df .dropna() .set_index('pl_name')
    target = tic_df.loc[[exoplanet]]
    tic_id = target['tic_id'] .values .tolist()[0]
    return tic_id

def _pl(tic_id):
    '''
    Returns a dataframe of all planet names and tic ID's.

    Returns
    -------
    Planet name corresponding to tic_id
    '''
    df = exo_nasa[['pl_name','tic_id']].dropna() \
        .sort_values('pl_name')
    df['tic_id'] = df['tic_id'].str[3:]
    df['tic_id'] = df['tic_id'].astype(int)
    df = df.set_index('tic_id')
    try:
        target = df.loc[[tic_id]]
    except KeyError:
        target = ''
    try:
        exoplanet = target['pl_name'] .values .tolist()[0]
    except Exception:
        exoplanet = ''
    return exoplanet


def _lc(exoplanet, mast, fast=False):
    '''
    # https://archive.stsci.edu/tess/bulk_downloads/bulk_downloads_ffi-tp-lc-dv.html
    out = DataFrame()
    for i in range(1,32):
        a = read_csv(f'TESS_curl/tesscurl_sector_{str(i)}_lc.sh')
        a['links'] = a['#!/bin/sh'].str.split().str[-1]
        a = a.drop(['#!/bin/sh'], axis=1)
        out = out.append(a)
    out.to_csv('MAST_lc.csv.gz', index=False)
    '''
    tic_id = _tic(exoplanet).replace('TIC ', '')
    lc_links = mast[mast['links'].str.contains(tic_id)] .values .tolist()
    lc_list = [i for j in lc_links for i in j]
    if fast==True:
        lc_test = [int(i[-35:-20]) for j in lc_links for i in j]
    else:
        lc_test = [int(i[-30:-15]) for j in lc_links for i in j]
    lc_links = []
    for i in range(len(lc_test)):
        if int(tic_id) == lc_test[i]:
            lc_links.append(lc_list[i])
    return lc_links, tic_id


def tess_candidate():
    '''
    Finds all candidate tic_ids and groups by amount of sector observations.

    '''
    _load_csv()
    here = os.path.dirname(os.path.abspath(__file__))
    mast_csv = f'{here}/data/Search/TESS_lc.csv.xz'
    mast = read_csv(mast_csv)
    for i in range(13,21):
        count = i
        df = mast.sort_values('links').reset_index()
        df = df.groupby(df.links.str[-30:-15].astype(int)).size().reset_index(name='Products')
        df = df.loc[df['Products']==count]
        df = df.rename(columns={'links':'TIC ID'})
        df['Exoplanet'] = df.apply(lambda row: _pl(row['TIC ID']), axis=1)
        here = os.path.dirname(os.path.abspath(__file__))
        os.makedirs(f'{here}/data/Candidates', exist_ok=True)
        df.to_csv(f'{here}/data/Candidates/{count}_sector_candidates.csv', index=False)
    

def _download_archive():
    os.makedirs('firefly/data', exist_ok=True)
    download_links = [ \
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # NASA
        # 'https://exoplanetarchive.ipac.caltech.edu/' +\
        # 'TAP/sync?query=select+' +\
        # 'pl_name,tic_id,pl_orbper,pl_orbsmax,pl_radj,pl_orbeccen,ttv_flag,' +\
        # 'st_teff,st_rad,st_mass,st_met,st_logg,pl_tranmid,pl_trandur,' +\
        # 'st_tefferr1,st_raderr1,st_meterr1,st_loggerr1,' +\
        # 'pl_orbincl,pl_orblper' +\
        # '+from+ps&format=csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # NASA COMP
        'https://exoplanetarchive.ipac.caltech.edu/' +\
        'TAP/sync?query=select+' +\
        'pl_name,tic_id,pl_orbper,pl_orbsmax,pl_radj,pl_bmasse,pl_bmassj,pl_orbeccen,ttv_flag,' +\
        'st_teff,st_rad,st_mass,st_met,st_logg,pl_tranmid,pl_trandur,' +\
        'st_tefferr1,st_raderr1,st_meterr1,st_loggerr1,' +\
        'pl_orbincl,pl_orblper,ra,dec,glat,glon,sy_dist,sy_plx,sy_tmag,' +\
        'pl_orbpererr1,pl_tranmiderr1,pl_orbsmaxerr1,pl_radjerr1,pl_orbinclerr1,' +\
        'pl_orblpererr1,pl_orbeccenerr1'
        '+from+pscomppars&format=csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # NASA Kepler Names
        'https://exoplanetarchive.ipac.caltech.edu/cgi-bin/' +\
        'nstedAPI/nph-nstedAPI?' +\
        'table=k2pandc',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # EU
        'http://exoplanet.eu/catalog/csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # OEC
        'https://raw.githubusercontent.com/OpenExoplanetCatalogue/' + \
        'oec_tables/master/comma_separated/open_exoplanet_catalogue.txt',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # ORG
        'http://exoplanets.org/csv-files/exoplanets.csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # TEP
        'https://www.astro.keele.ac.uk/jkt/tepcat/allplanets-csv.csv'
    ]
    archive = ['nasa', 'eu', 'kep', 'oec', 'org', 'tep']
    for i, download_link in enumerate(download_links):
        try:
            i = archive[i]
            csv = f'firefly/data/{i}.csv.gz'
            if not os.path.exists(csv):
                print(f'Caching the {i.upper()} Exoplanet Archive.')
                df = read_csv(download_link)
                df.to_csv(csv, index=False)
            seven_days_ago = datetime.now() - timedelta(days=7)
            filetime = datetime.fromtimestamp(os.path.getctime(csv))
            if filetime < seven_days_ago:
                print(f'{i.upper()} Archive is 7 days old. Updating.')
                df = read_csv(download_link)
                df.to_csv(csv, index=False)
            else:
                pass
        except Exception:
            print(f'Could not contact the server for the {i.upper()} archive.')
            pass


def _download_archive_threaded(download_link, i):
    os.makedirs('firefly/data', exist_ok=True)
    try:
        csv = f'firefly/data/{i}.csv.gz'
        if not os.path.exists(csv):
            print(f'Caching the {i.upper()} Exoplanet Archive.')
            df = read_csv(download_link)
            df.to_csv(csv, index=False)
        seven_days_ago = datetime.now() - timedelta(days=7)
        filetime = datetime.fromtimestamp(os.path.getctime(csv))
        if filetime < seven_days_ago:
            print(f'{i.upper()} Archive is 7 days old. Updating.')
            df = read_csv(download_link)
            df.to_csv(csv, index=False)
        else:
            pass
    except Exception:
        print(f'Could not contact the server for the {i.upper()} archive.')
        pass

def download_archive_threaded_execute():
    download_links = [ \
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # NASA
        # 'https://exoplanetarchive.ipac.caltech.edu/' +\
        # 'TAP/sync?query=select+' +\
        # 'pl_name,tic_id,pl_orbper,pl_orbsmax,pl_radj,pl_orbeccen,ttv_flag,' +\
        # 'st_teff,st_rad,st_mass,st_met,st_logg,pl_tranmid,pl_trandur,' +\
        # 'st_tefferr1,st_raderr1,st_meterr1,st_loggerr1,' +\
        # 'pl_orbincl,pl_orblper' +\
        # '+from+ps&format=csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # NASA COMP
        'https://exoplanetarchive.ipac.caltech.edu/' +\
        'TAP/sync?query=select+' +\
        'pl_name,tic_id,pl_orbper,pl_orbsmax,pl_radj,pl_bmasse,pl_bmassj,pl_orbeccen,ttv_flag,' +\
        'st_teff,st_rad,st_mass,st_met,st_logg,pl_tranmid,pl_trandur,' +\
        'st_tefferr1,st_raderr1,st_meterr1,st_loggerr1,' +\
        'pl_orbincl,pl_orblper,ra,dec,glat,glon,sy_dist,sy_plx,sy_tmag,' +\
        'pl_orbpererr1,pl_tranmiderr1,pl_orbsmaxerr1,pl_radjerr1,pl_orbinclerr1,' +\
        'pl_orblpererr1,pl_orbeccenerr1'
        '+from+pscomppars&format=csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # NASA Kepler Names
        'https://exoplanetarchive.ipac.caltech.edu/' +\
        'TAP/sync?query=select+*+from+k2pandc&format=csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # EU
        'http://exoplanet.eu/catalog/csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # OEC
        'https://raw.githubusercontent.com/OpenExoplanetCatalogue/' + \
        'oec_tables/master/comma_separated/open_exoplanet_catalogue.txt',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # ORG
        'http://exoplanets.org/csv-files/exoplanets.csv',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # TEP
        'https://www.astro.keele.ac.uk/jkt/tepcat/allplanets-csv.csv'
    ]
    archive = ['nasa', 'kep', 'eu', 'oec', 'org', 'tep']
    threads = []
    for i in range(5):
        j = archive[i]
        download_link = download_links[i]
        t = threading.Thread(target=_download_archive_threaded, args=(download_link, j))
        t.daemon = True
        threads.append(t)
    for i in range(5):
        threads[i].start()
    for i in range(5):    
        threads[i].join()


def _nasa_full():
    '''
    Downloads the entire NASA Exoplanet Archive

    Returns
    -------
    data/nasa_full.csv

    '''
    download_link =  \
    'https://exoplanetarchive.ipac.caltech.edu/' +\
    'TAP/sync?query=select+*+from+ps&format=csv'
    df = read_csv(download_link)
    df.to_csv('firefly/data/nasa_full.csv.xz', index=False)


def estimate_t14(Rp, Rs, a, P):
    '''
    Estimates t14 in minutes, if P is in days
    '''
    AU = 1.495978707e11
    R_sun = 6.957e8
    R_jup = 71492000
    return (Rp * R_jup + Rs * R_sun)/(np.pi * a * AU) * P * 24 * 60


def _IQR(df, sigma=1):
    # Only keep IQR of data
    if sigma==1:
        Q1 = df.quantile(0.45)
        Q3 = df.quantile(0.55)
    if sigma==2:
        Q1 = df.quantile(0.35)
        Q3 = df.quantile(0.65)
    if sigma==3:
        Q1 = df.quantile(0.25)
        Q3 = df.quantile(0.75)
    IQR = Q3 - Q1
    trueList = ~((df < (Q1 - 1.5 * IQR)) | (df > (Q3 + 1.5 * IQR)))
    return df[trueList]
  
    
def priors(exoplanet, archive='eu', save=False, user=True, auto=True, fit_ttv=False):
    '''
    Generates priors from 4 exoplanet archives, nasa, eu, oec and exoplanets.org

    Parameters
    ----------
    exoplanet : str, 'wasp43b'
        The exoplanet to create a prior object for.
    archive : str, {'eu', 'nasa', 'org', 'all'}
        The archive to generate priors from. All takes the IQR of all
        archives (including OEC) and then the mean.
        The default is 'eu'.
    save : bool, optional
        Internal use only. The default is False.
    user : bool, optional
        Internal use only. The default is True.

    '''
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Download NASA archive
    #_download_archive()
    if user==True:
        download_archive_threaded_execute()
        _load_csv()
        highest, ratios = _search_all(exoplanet)
        exoplanet = highest[0]
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    if fit_ttv==True:
        archive = 'spearnet'
    archive_list = [exo_nasa, exo_eu, exo_org, exo_oec, exo_spearnet, exo_spearnet_uncoup]
    if archive=='nasa':
        exo_archive = archive_list[0].set_index('pl_name')
    elif archive=='eu':
        exo_archive = archive_list[1].set_index('pl_name')
    elif archive=='org':
        exo_archive = archive_list[2].set_index('pl_name')
    elif archive=='spearnet':
        exo_archive = archive_list[4].set_index('pl_name')
    elif archive=='all':
        exo_archive = concat(archive_list).set_index('pl_name')
    if archive =='all':
        exo_archive = concat(archive_list)
        exo_archive = exo_archive[exo_archive['pl_name'].notna()]
        exo_archive['pl_name'] = \
            Categorical(exo_archive['pl_name'],
            ordered=True,
            categories=natsorted(exo_archive['pl_name'].unique()))
        exo_archive = exo_archive.sort_values('pl_name').set_index('pl_name') \
                      .drop(['Archive'], axis=1)
        # exo = exo_archive.mean(axis=0, level=0)
        exo_archive.to_csv('firefly/data/all.csv.gz')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Pick Out Chosen Exoplanet Priors
    try:
        choice = exo_archive.loc[[exoplanet]]
    except KeyError:
        sys.exit('The chosen target is either spelt incorrectly, or does not '
                 f'exist in the {archive.upper()} archive.')
    try:
        # tic = df['tic_id'] .drop_duplicates() .dropna()[0]
        tic = _tic(exoplanet)
    except IndexError:
        tic = 'N/A'
    # Choose Archive
    if archive=='spearnet':
        exo_limb = archive_list[0].set_index('pl_name')
        choice_limb = exo_limb.loc[[exoplanet]]
        s_limb = choice_limb.mean()
    # Turn into a series
    s = choice.iloc[0]
    t0 = s.loc['pl_tranmid']
    if archive=='all':
        t0 = choice['pl_tranmid'].max()
        df = _IQR(choice, sigma=3)
        s = df.mean()
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Values for calculation
    P = s.loc['pl_orbper']
    a = s.loc['pl_orbsmax']
    i = s.loc['pl_orbincl']
    w = s.loc['pl_orblper']
    ecc = s.loc['pl_orbeccen']
    rp = s.loc['pl_radj']
    # rserr = s.loc['st_raderr1']
    rs = s.loc['st_rad']
    z = s.loc['st_met']
    zerr =  s.loc['st_meterr1']
    ms = s.loc['st_mass']
    T = s.loc['st_teff']
    # Terr =  s.loc['st_tefferr1']
    t14 = s.loc['pl_trandur'] * 60
    logg = s.loc['st_logg']
    # loggerr = s.loc['st_loggerr1']
    if archive=='nasa':
        ra = s.loc['ra']
        dec = s.loc['dec']
        dist = s.loc['sy_dist']
        t_mag = s.loc['sy_tmag']
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Assign Host data to Transitfit
    host_T = (T, T * 5e-2)
    host_z = (z, zerr)
    host_r = (rs, rs * 5e-2)
    host_logg = (logg, logg * 5e-2)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Nan checks
    # G = 6.67408e-11
    # AU = 1.495978707e11
    # sol = 1.98847e30
    # P
    # if (np.isnan(P) and not np.isnan(a)):
    #     # Added small correction factor to star mass
    #     P = 2 * np.pi * np.sqrt((a * AU)**3 /
    #                             (G * 0.963 * ms * sol)) / (60 * 60 * 24)
    # # a
    # elif (np.isnan(a) and not np.isnan(P)):
    #     a = (((P * 24 * 60 * 60)**2 * G * ms * sol /
    #           (4 * np.pi**2))**(1 / 3)) / AU
    # w
    if archive=='spearnet':
        Perr = s.loc['pl_orbpererr1']
        aerr = s.loc['pl_orbsmaxerr1']
        ierr = s.loc['pl_orbinclerr1']
        werr = s.loc['pl_orblpererr1']
        eccerr = s.loc['pl_orbeccenerr1']
        t0err = s.loc['pl_tranmiderr1']
        rperr = s.loc['pl_radjerr1']
        rs = s_limb.loc['st_rad']
        z = s_limb.loc['st_met']
        zerr =  s_limb.loc['st_meterr1']
        ms = s_limb.loc['st_mass']
        T = s_limb.loc['st_teff']
        logg = s_limb.loc['st_logg']
        host_T = (T, T * 5e-2)
        host_z = (z, zerr)
        host_r = (rs, rs * 5e-2)
        host_logg = (logg, logg * 5e-2)
        t14 = estimate_t14(rp, rs, a, P)
    if (np.isnan(w) or w==0):
        w = 90
    # ecc
    if np.isnan(ecc):
        ecc = 0
    # t14
    if np.isnan(t14):
        t14 = estimate_t14(rp, rs, a, P)
    # logg
    if np.isnan(logg):
        logg, err_logg = calculate_logg((ms, ms * 1e-2), (rs, rs * 1e-2))
        host_logg = (logg, err_logg)
    # z
    if (np.isnan(zerr) and z!=0):
        host_z = (z, 0.05)
    if (np.isnan(z) or z==0):
        host_z = (0, 0.05)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Assign Exoplanet Priors to TransitFit
    radius_const = 0.1027626851
    cols = [['P', ['fixed' if archive=='spearnet' else 'gaussian'][0], P,
                ['' if archive=='spearnet' else P * 1e-3][0], ''],
            ['t0', 'gaussian', t0, [7e-3 if archive=='spearnet' else 7e-3][0], ''],
            ['a', 'gaussian', a, [a * 2e-1 if archive=='spearnet' else a * 2e-1][0], ''],
            ['inc', 'gaussian', i, [i * 2e-1 if archive=='spearnet' else i * 2e-1][0], ''],
            ['w', ['fixed' if (w==90 or ecc==0) else 'gaussian'][0],
             [90 if ecc==0 else w][0],
             ['' if (w==90 or ecc==0)
              else (360-np.abs(w))/2 if 180 < w < 270
              else (360-np.abs(w)) if 270 <= w < 360
              else werr if archive=='spearnet'
              else w * 0.5][0], ''],
            ['ecc', ['fixed' if (ecc==0 or w==90)
                     else 'gaussian'][0], [0 if w==90 else ecc][0],
             ['' if (ecc==0 or w==90)
              else ecc * 0.5 if archive=='spearnet'
              else ecc * 0.5][0], ''],
            ['rp', ['uniform' if archive=='spearnet' else 'uniform'][0],
             [0.5*rp if archive=='spearnet' else (0.5 * radius_const * rp / rs)][0],
             [1.5*rp if archive=='spearnet' else (1.5 * radius_const * rp / rs)][0], 0]]
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Save the priors
    priors_csv = DataFrame(cols, columns=['Parameter', 'Distribution',
                                          'Input_A', 'Input_B', 'Filter'])
    if save == True:
        priors = f'firefly/{exoplanet}/{exoplanet} Priors.csv'
        if not os.path.exists(priors):
            priors_csv.to_csv(priors, index=False, header=True)
        else:
            pass
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # For printing variables only
    cols = [['P', ['fixed' if archive=='spearnet' else 'gaussian'][0], P,
                ['' if archive=='spearnet' else P * 1e-3][0], ''],
            ['t0', 'gaussian', t0, [7e-3 if archive=='spearnet' else 7e-3][0], ''],
            ['a', 'gaussian', a, [a * 2e-1 if archive=='spearnet' else a * 2e-1][0], ''],
            ['inc', 'gaussian', i, [i * 2e-1 if archive=='spearnet' else i * 2e-1][0], ''],
            ['w', ['fixed' if (w==90 or ecc==0) else 'gaussian'][0],
             [90 if ecc==0 else w][0],
             ['' if (w==90 or ecc==0)
              else (360-np.abs(w))/2 if 180 < w < 270
              else (360-np.abs(w)) if 270 <= w < 360
              else werr if archive=='spearnet'
              else w * 0.5][0], ''],
            ['ecc', ['fixed' if (ecc==0 or w==90)
                     else 'gaussian'][0], [0 if w==90 else ecc][0],
             ['' if (ecc==0 or w==90)
              else ecc * 0.5 if archive=='spearnet'
              else ecc * 0.5][0], ''],
            ['rp', ['uniform' if archive=='spearnet' else 'uniform'][0],
             [0.5*rp if archive=='spearnet' else (0.5 * radius_const * rp / rs)][0],
             [1.5*rp if archive=='spearnet' else (1.5 * radius_const * rp / rs)][0], 0],
            ['t14', '', t14, '', ''],
            ['host_T', '', host_T[0], host_T[1], ''],
            ['host_z', '', host_z[0], host_z[1], ''],
            ['host_r', '', host_r[0], host_r[1], ''],
            ['host_logg', '', host_logg[0], host_logg[1], '']]
    repack = DataFrame(cols, columns=['Parameter', 'Distribution',
                                      'Input A', 'Input B', 'Filter'])
    is_nan = repack.isnull().values.any()
    #if (is_nan and user==False and auto==True):
     #   raise NaNError(f'Skipping {exoplanet} due to missing prior data.')
    if archive=='all':
        print(f'\nPriors generated from the NASA, EU, OEC and ORG Archives for'
              f' {exoplanet} ({tic}).\n')
    elif archive not in ['nasa', 'eu', 'org', 'spearnet']:
        archive = 'NASA'
    else:
        print(f'\nPriors generated from the {archive.upper()} Archive for'
              f' {exoplanet} ({tic}).\n')
    print(tabulate(repack, tablefmt='psql', showindex=False, headers='keys'))
    if (archive=='nasa' and user==False):
        return host_T, host_z, host_r, host_logg, t0, P, t14, repack, ra, dec, dist, t_mag
    if user==False:
        return host_T, host_z, host_r, host_logg, t0, P, t14, repack


def tess(archive='eu', survey=None):
    '''
    Currently there are 420 tess targets with full prior and lightcurve sets.

    Parameters
    ----------
    k : int, optional
        Returns a random set of targets. The default is 10.
    survey : str, optional
        Choice of survey to filter by. The default is None.

    Returns
    -------
    targets : str
        Viable tess target list.

    '''
    here = os.path.dirname(os.path.abspath(__file__))
    tess = f'{here}/data/Targets/{archive}_tess_viable.csv'
    tess_ttv = f'{here}/data/Targets/tess_ttv_viable.csv'
    targets = read_csv(tess)['Exoplanet']
    ttv_targets = read_csv(tess_ttv)['Exoplanet']
    if survey != None:
        targets = [s for s in targets if survey in s]
        ttv_targets = [s for s in ttv_targets if survey in s]
    all_targets = natsorted(targets)
    ttv_targets = natsorted(ttv_targets)
    return targets, all_targets, ttv_targets


def gen_tess(archive='nasa', cadence=120):
    '''Generates all TESS targets which have a fitsfile.'''

    def count_products(tic_id):
        '''Counts the amount of TESS fitsfiles on MAST.'''
        time.sleep(0.01)
        search = obs.query_criteria(dataproduct_type=['timeseries'],
                                    project='TESS',
                                    provenance_name=['SPOC'],
                                    t_exptime=cadence,
                                    target_name=tic_id).to_pandas()
        search = search[search['t_exptime']==cadence]
        search = search[~search.dataURL.str.endswith('_dvt.fits')]
        if len(search) is None:
            return 0
        return len(search)

    _load_csv()
    if archive=='nasa':
        df = exo_nasa
    elif archive=='eu':
        df = exo_eu
    elif archive=='nasa_full':
        download_link =  \
        'https://exoplanetarchive.ipac.caltech.edu/' +\
        'TAP/sync?query=select+*+from+ps&format=csv'
        print('Downloading the full NASA archive.')
        df = read_csv(download_link)
        print('Download complete.')
    print('Generating product list and transit counts.')
    print('The process should take no longer than 15 minutes.')
    df['tic_id'] = df['tic_id'].str.replace('TIC ', '')
    df[f'Products ({cadence} Cadence)'] = df['tic_id'].progress_apply(count_products)
    df['Transits'] = np.ceil((0.8 * 27.4 / df['pl_orbper']) * df[f'Products ({cadence} Cadence)'])
    df['Transits'] = df['Transits'].fillna(-1).astype(int)
    df['pl_name'] = \
            Categorical(df['pl_name'],
            ordered=True,
            categories=natsorted(df['pl_name'].unique()))
    df = df.sort_values('pl_name')
    here = os.path.dirname(os.path.abspath(__file__))
    df.to_csv(f'{here}/data/Targets/ML_nasa_tess_viable.csv.xz', index=False)
    df.to_csv('firefly/data/ML_nasa_tess_viable.csv.xz', index=False)
    print('Operation completed.')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb  6 13:14:20 2021

@author: Steven Charles-Mindoza
"""

import pandas as pd
import os
from matplotlib import pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.colors import to_rgb
from matplotlib import colors
from astropy.timeseries import LombScargle
from astropy import units as u
import numpy as np
import matplotlib as mpl
import seaborn as sns
mpl.rcParams['figure.dpi'] = 300

base_context = {
                "font.size": 18,
                "axes.labelsize": 18,
                "axes.titlesize": 14,
                "xtick.labelsize": 16,
                "ytick.labelsize": 16,
                "legend.fontsize": 16,

                "axes.linewidth": 1.25,
                "grid.linewidth": 1,
                "lines.linewidth": 1.5,
                "lines.markersize": 6,
                "patch.linewidth": 1,

                "xtick.major.width": 0,
                "ytick.major.width": 0,
                "xtick.minor.width": 0,
                "ytick.minor.width": 0,

                "xtick.major.size": 0,
                "ytick.major.size": 0,
                "xtick.minor.size": 0,
                "ytick.minor.size": 0,

                }

sns.set_theme(style="whitegrid")
sns.set_context('paper')
sns.set_palette('deep')

def change_width(ax, new_value):
    for patch in ax.patches:
        current_width = patch.get_width()
        diff = current_width - new_value
        patch.set_width(new_value)
        patch.set_x(patch.get_x() + diff * .5)


def plot_epoch(sub=False):
    sns.set_theme(style="whitegrid", rc = base_context)
    if sub==True:
        fig, ax = plt.subplots(figsize=(20,25))
    archive_list = ['eu', 'nasa', 'org', 'all']
    for i, archive in enumerate(archive_list):
        here = os.path.dirname(os.path.abspath(__file__))
        df = pd.read_csv(f'{here}/data/Targets/{archive}_tess_viable.csv') \
        .sort_values('Epochs', ascending=False).reset_index(drop=True)
        # df = pd.read_csv(f'Targets/{archive}_tess_viable.csv') \
        # .sort_values('Epochs', ascending=False).reset_index(drop=True)
        # candidates = f'{len(df)} {archive.upper()} Archive ' +\
        #            'Candidates Ranked by Epoch Count ' +\
        #             f"({df['Epochs'].sum()} Total)"
        candidates = f'{len(df)} {archive.upper()} Archive ' +\
             'Candidates Ranked by Descending Observed Transits'
        df[candidates] = 1
        df['Candidate Cumsum'] = df[candidates].cumsum()
        total_candidates = df[candidates].sum()
        df['Candidate Frequency'] = df['Candidate Cumsum'] / total_candidates
        
        df['Epoch Cumsum'] = df['Epochs'].cumsum()
        total_epochs = df['Epochs'].sum()
        df['Epoch Frequency'] = df['Epoch Cumsum'] / total_epochs
        # highlights = df.groupby('Frequency')['Exoplanet', 'Epochs']
        # .agg({'Exoplanet':'count','Epochs':'sum'})

        highlights = df.groupby(pd.cut(df["Epoch Frequency"],
                                       np.arange(0, 1.0+0.1, 0.1))).sum()
        cumsum_cand = highlights[candidates].cumsum()
        cumsum_epochs = highlights['Epochs'].cumsum()
        if sub==False:
            fig, ax = plt.subplots(figsize=(15,45))
            sns.despine(fig=fig, ax=ax)
        temp = 411 + i
        ax=plt.subplot(temp)
        [ax.axvline(x=i+0.4, color='k', marker=',',
                    alpha=0.5, ymin=0,ymax=0.15) for i in range(10)]
        [ax.axvline(x=i+0.45, color='k', marker=',',
                    alpha=0.5, ymin=0.8,ymax=1) for i in range(10)]
        # [ax.text(i+0.34,total_epochs/47,
        #          f"{i+1}0% - {cumsum_cand[i]} Candidates - {cumsum_epochs[i]} Transits",rotation=90)
        #          for i in range(10)]
        [ax.text(i+0.34,total_epochs/47,
                 f"{i+1}0% of All Transits - {round(cumsum_cand[i]*100/total_candidates)}% Candidates",rotation=90)
                 for i in range(10)]
        [ax.text(i-0.1, highlights['Epochs'][i] + 5,
                 str(highlights[candidates][i]),rotation=0) for i in range(10)]
        # Top Planets
        twenty_perc = cumsum_cand[2]
        textstr = '\n'.join(df['Exoplanet'][0:twenty_perc])
        props = dict(boxstyle='round', facecolor='white', alpha=0.1)
        ax.text(1.075, 0.96, f'30% of \nAll Transits \n{twenty_perc} Candidates',
                transform=ax.transAxes, fontsize=16, weight='bold',
                verticalalignment='top', bbox=props, ha='center')
        ax.text(1.02, 0.85, textstr, transform=ax.transAxes, fontsize=16,
                verticalalignment='top', bbox=props)
        # PLOT!
        highlights['Observed Transits'] = highlights['Epochs']
        sns.barplot(ax=ax, data=highlights,
                    x=candidates, y = 'Observed Transits',
                    dodge=False, palette='rocket')
        change_width(ax, 0.6)
        # ax.get_legend().remove()
        
        ax.xaxis.tick_bottom()
        column_labels = [f'{i}0 - {i+1}0%'.replace('00 - 10%', '0 - 10%') for i in range(10)]
        ax.set_xticklabels(column_labels, minor=False)
        if sub ==False:
            plt.savefig(f'{archive}_epoch_rank.jpg', bbox_inches='tight')
    if sub==True:
        plt.savefig('epoch_rank.jpg', bbox_inches='tight')
        

def lc_plot(file, flatten=False):
    sns.set_theme(style="whitegrid", rc = base_context)
    from lightkurve import LightCurve
    df = pd.read_csv(file).dropna()
    time = df['Time'].values - 2457000
    flux = df['Flux'].values
    flux_err = df['Flux err'].values
    
    lc = LightCurve(time, flux, flux_err, time_format='btjd')
    lc = lc.remove_outliers().normalize()
    if flatten==True:
        lc = lc.remove_outliers().flatten()
    
    pg = lc.to_periodogram(oversample_factor=25)
    period = float(pg.period_at_max_power/u.d)
    pg = lc.to_periodogram(oversample_factor=25, maximum_period=period*1.25,
                           minimum_period=period*0.75)
    # PLOT!
    fig, ax = plt.subplots(figsize=(15,15))
    ax=plt.subplot(311)
    plt.errorbar(lc.time, lc.flux,
                 lc.flux_err, color='b',
                 alpha=0.2, zorder=1, capsize=2, ls='none')
    plt.scatter(lc.time, lc.flux, color='k', s=0.5, alpha=0.6, zorder=2)
    plt.xlabel('Time (BTJD)')
    plt.ylabel('Flux')
    
    ax=plt.subplot(312)
    plt.plot(pg.period, pg.power, color='k', alpha=0.8)
    plt.xlabel('Period')
    plt.ylabel('Power')
    #pg.show_properties()

    # Folded
    period = pg.period_at_max_power
    condition = (lc.flux==np.min(lc.flux))
    t0 = np.where(condition)[0][0]
    t0 = lc.time[t0]
    folded_lc = lc.fold(period=period, t0=t0)
    
    ax=plt.subplot(313)
    plt.errorbar(folded_lc.time, folded_lc.flux,
                 folded_lc.flux_err, color='b',
                 alpha=0.2, zorder=1, capsize=2, ls='none')
    plt.scatter(folded_lc.time, folded_lc.flux,
                alpha=0.4, color='k', zorder=2, s=2)
    plt.xlabel('Phase')
    plt.ylabel('Flux')
    filename = f'{file}'.replace('.csv', '')
    plt.savefig(f'{filename}.jpg', bbox_inches='tight')


def mw():
    from astropy import units as u
    import astropy.coordinates as apycoords
    from mw_plot import MWSkyMap, MWPlot
    here = os.path.dirname(os.path.abspath(__file__))
    # Viable Targets
    df = pd.read_csv(f'{here}data/Targets/nasa_tess_viable.csv')
    ra = df['RA'].values * u.deg
    dec = df['DEC'].values * u.deg
    z = df['Epochs'].values
    distance = df['Distance'] .values * u.pc
    c = apycoords.SkyCoord(ra=ra, dec=dec, distance=distance, frame='icrs')
    
    # All Targets
    df2 = pd.read_csv('firefly/data/nasa.csv.gz')
    ra2 = df2['ra'].values * u.deg
    dec2 = df2['dec'].values * u.deg
    distance2 = df2['sy_dist'] .values * u.pc
    c2 = apycoords.SkyCoord(ra=ra2, dec=dec2, distance=distance2, frame='icrs')
    
    plot_instance = MWPlot(mode='face-on', center=(0, 0)*u.kpc, radius= 12*u.kpc,
                       unit=u.kpc, coord='galactic', annotation=True,  grayscale=False)
    plot_instance.fontsize = 35  # fontsize for matplotlib plotting
    plot_instance.figsize = (20, 20)  # figsize for matplotlib plotting
    plot_instance.dpi = 300  # dpi for matplotlib plotting
    plot_instance.cmap = 'hot'  # matplotlib cmap: https://matplotlib.org/examples/color/colormaps_reference.html
    #plot_instance.clim = (vmin, vmax) # colorbar range
    plot_instance.imalpha = 1  # alpha value for the milkyway image
    plot_instance.s = 100.0  # make the scatter points bigger
    plot_instance.tight_layout = True # whether plt.tight_layout() will be run
    

    plot_instance.mw_scatter(-c.galactic.cartesian.x,
                             c.galactic.cartesian.y, [z, 'Transits Observed'])
    # plot_instance.mw_scatter(-c2.galactic.cartesian.x,
    #                          c2.galactic.cartesian.y, [distance2, ''] )
    plot_instance.savefig('mw_zoom_out.png')
    
    
    plot_instance = MWPlot(mode='face-on', center=(0, 0)*u.kpc, radius= 2*u.kpc,
                       unit=u.kpc, coord='galactic', annotation=True,  grayscale=False)
    plot_instance.fontsize = 35  # fontsize for matplotlib plotting
    plot_instance.figsize = (20, 20)  # figsize for matplotlib plotting
    plot_instance.dpi = 300  # dpi for matplotlib plotting
    plot_instance.cmap = 'hot'  # matplotlib cmap: https://matplotlib.org/examples/color/colormaps_reference.html
    #plot_instance.clim = (vmin, vmax) # colorbar range
    plot_instance.imalpha = 1  # alpha value for the milkyway image
    plot_instance.s = 100.0  # make the scatter points bigger
    plot_instance.tight_layout = True # whether plt.tight_layout() will be run
    
    plot_instance.scatter(-c2.galactic.cartesian.x,
                             c2.galactic.cartesian.y, color='dimgrey', zorder=1)
    plot_instance.mw_scatter(-c.galactic.cartesian.x,
                             c.galactic.cartesian.y, [z, 'Transits Observed'],
                             zorder=2)
    plot_instance.savefig('mw.png')
    
    
    plot_instance = MWPlot(mode='face-on', center=(0, 0)*u.kpc, radius= 1*u.kpc,
                       unit=u.kpc, coord='galactic', annotation=True,  grayscale=False)
    plot_instance.fontsize = 35  # fontsize for matplotlib plotting
    plot_instance.figsize = (20, 20)  # figsize for matplotlib plotting
    plot_instance.dpi = 300  # dpi for matplotlib plotting
    plot_instance.cmap = 'hot'  # matplotlib cmap: https://matplotlib.org/examples/color/colormaps_reference.html
    #plot_instance.clim = (vmin, vmax) # colorbar range
    plot_instance.imalpha = 1  # alpha value for the milkyway image
    plot_instance.s = 100.0  # make the scatter points bigger
    plot_instance.tight_layout = True # whether plt.tight_layout() will be run
    
    plot_instance.scatter(-c2.galactic.cartesian.x,
                             c2.galactic.cartesian.y, color='dimgrey', zorder=1)
    plot_instance.mw_scatter(-c.galactic.cartesian.x,
                             c.galactic.cartesian.y, [z, 'Transits Observed'],
                             zorder=2)
    plot_instance.savefig('mw_zoom_in.png')
    

def oc(t0, t0_err, file='Complete_Results.csv', exoplanet=None):
    path_ttv = file
    data_ttv = pd.read_csv(path_ttv).set_index('Parameter') \
                .filter(like = 't0', axis=0).drop(['Telescope', 'Filter'], axis=1)
    
    t0_o = t0 # data['Best'] .values
    t0_oerr = t0_err # data['Error'].astype(float) .values
    t0_c = data_ttv['Best'] .values
    t0_cerr = data_ttv['Error'].astype(float) .values
    epoch_no = (data_ttv['Epoch'].astype(int) + 1) . values
    
    ominusc = t0_c - t0_o
    ominuscerr = t0_cerr - t0_oerr
    ominusc *= 24 * 60
    ominuscerr *= 24 * 60
    from sklearn.preprocessing import scale
    ominusc = scale(ominusc)

    
    # Generalised Lomb Scargle Floating Mean
    ls = LombScargle(epoch_no, ominusc, ominuscerr)
    frequency, power = ls.autopower(nyquist_factor=1, samples_per_peak=100)
    
    # False Alarm Probability
    fap = ls.false_alarm_probability(power.max())
    levels = [0.99, 0.5, 0.05, 0.01]
    false_alarm_levels = ls.false_alarm_level(levels)
    pack = {'FAP':false_alarm_levels, 'Percentage':levels}
    period = 1/frequency[np.argmax(power)]
    # Get the best frequency for plotting??
    fit_x = np.linspace(epoch_no.min(), epoch_no.max(), 1000)
    fit_y = ls.model(fit_x, frequency[np.argmax(power)])
    
    
    fig, ax = plt.subplots(figsize=(15,15))
    
    ax=plt.subplot(311)
        
    # Make figure and axes
    fig = plt.figure(figsize=(12,8))
    gs = gridspec.GridSpec(2, 1)

    oc_ax = fig.add_subplot(gs[0])
    ls_ax = fig.add_subplot(gs[1])

    #Plot data
    oc_ax.errorbar(epoch_no, ominusc, ominuscerr, marker='.',
                   elinewidth=0.8, color='dimgrey', linestyle='',
                   capsize=2, alpha=0.8, zorder=1)
    oc_ax.scatter(epoch_no, ominusc, marker='.', color='dimgrey', zorder=2)
    oc_ax.axhline(0, color='black', linestyle='--', linewidth=1)
    oc_ax.plot(fit_x, fit_y, color='red', alpha=0.8)
    
    #ls_ax.scatter(1/frequency, power, color='dimgrey', alpha=0.1, zorder=1, s=10)
    ls_ax.plot(1/frequency, power, linestyle='-', linewidth=0.75, zorder=2,
               marker='', color='k')
    pos = period * 1.97
    for (level, i) in zip(false_alarm_levels, levels):
        ls_ax.axhline(level, color='r', linestyle='--', alpha=0.8)
        ls_ax.annotate(f'{str(int(i*100))}'+'$\%$',
                       (pos, level*1.018), color='k', ha='center')
    ls_ax.annotate(f'Period: {int(period)}\n{fap*100:.2f}$\%$',
                        (period, power.max()*1.03), color='k', weight='bold', ha='center')
    # Sort out labels etc
    oc_ax.set_xlabel('Epoch')
    oc_ax.set_ylabel('O-C (minutes)')

    ls_ax.set_xlabel('Period (Epochs)')
    ls_ax.set_ylabel('Power')

    oc_ax.tick_params('both', which='both', direction='in', bottom=True, left=True)
    ls_ax.tick_params('both', which='both', direction='in', bottom=True, left=True)

    ls_ax.set_xscale('linear')
    ls_ax.set_xlim([0, period * 2])
    upper_y_fap = false_alarm_levels[2] * 1.5
    upper_y_pow = max(power) * 1.2
    upper_y = np.max([upper_y_fap, upper_y_pow])
    ls_ax.set_ylim([-0.01, upper_y])
    fig.tight_layout()
    if exoplanet==None:
        fig.savefig('O-C.jpg', bbox_inches='tight')
    else:
        fig.savefig(f"firefly/{exoplanet}/{exoplanet.lower().replace(' ', '')}_o-c.jpg",
                    bbox_inches='tight')

    
def make_epoch_converter(exoplanet):
    '''
    Converts between the epoch idx used in TF and the epoch idx used in
    file naming (they don't match because we removed some things)

    Returns dict with {epoch_idx : filename}
    '''
    data = pd.read_csv(f'firefly/{exoplanet}/data_paths.csv')
    #print(data.columns)
    converter = {}

    for i, row in data.iterrows():
        #print(row[' Epoch'])
        converter[row['Epochs']] = row['Path']#[15:]

    return converter

def find_epoch_no(epoch_idx, t0, P, epoch_converter):
    '''
    Given an epoch idx from TransitFit, finds the number of orbits that have
    passed to get to this epoch - Useful for things like O-C plots!
    '''
    # Load up the relevant data file
    path = epoch_converter[epoch_idx]

    data = pd.read_csv(path)

    # Get the last time value and calculate how many periods have elapsed between t0 and then.
    return (data['Time'].values[-1] - t0)// P
  
def oc_fold(t0, t0err, P, file='Complete_results.csv', exoplanet=None, longterm=True):
    '''
    Loads in the t0 and errors
    '''
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Read in TTV Data
    converter = make_epoch_converter(exoplanet)
    path_ttv = file
    data_ttv = pd.read_csv(path_ttv).set_index('Parameter') \
                .filter(like = 't0', axis=0).drop(['Telescope', 'Filter'], axis=1)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Compute o-c
    t0_o = t0
    t0_oerr = t0err
    t0_c = data_ttv['Best'] .values
    t0_cerr = data_ttv['Error'].astype(float) .values
    epoch_no = np.array([find_epoch_no(i, t0_o, P, converter)
                         for i in range(len(t0_c))]) * P
    epoch_no = epoch_no - epoch_no[0]
    ominusc = t0_o  - t0_c
    ominuscerr = t0_cerr
    ominusc *= 24 * 60
    ominuscerr *= 24 * 60
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Chi 2
    from sklearn.metrics import mean_absolute_error
    loss = mean_absolute_error(ominusc,ominuscerr)
    dof = len(ominusc)
    chi2_red = np.sum( (ominusc)**2 / (t0_cerr)**2)/dof
    sigma = np.sqrt(2./len(ominusc))
    nsig = (chi2_red-1)/sigma
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Do the Lomb-Scargel stuff.
    ls = LombScargle(epoch_no, ominusc, ominuscerr)
                
    max_p = P * len(ominusc)//2
    if longterm==True:
        frequency, power = ls.autopower(nyquist_factor=1, samples_per_peak=75)
    else:
        frequency, power = ls.autopower(nyquist_factor=1, samples_per_peak=25,
                                    minimum_frequency=1/max_p)
    best_f = frequency[np.argmax(power)]
    best_P = 1/frequency[np.argmax(power)]
    epoch_phase = (epoch_no - (epoch_no //best_P) * best_P)/best_P
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # FAP
    fap = ls.false_alarm_probability(power.max())
    levels = [0.5, 0.05, 0.01]
    false_alarm_levels = ls.false_alarm_level(levels)
    pack = {'FAP':false_alarm_levels, 'Percentage':levels}
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Fold and Fits
    epoch_phase = (epoch_no  - (epoch_no //best_P) * best_P)/best_P
    fit_x = np.linspace(epoch_no.min(), epoch_no.max(), 10000)
    fit_y = ls.model(fit_x, frequency[np.argmax(power)])
    fit_x_phase = (fit_x - (fit_x //best_P) * best_P)/best_P
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Begin Plots
    fig = plt.figure(figsize=(12,8))
    gs = gridspec.GridSpec(3, 1)
    plt.set_cmap('plasma')
    
    oc_ax = fig.add_subplot(gs[0])
    phase_ax = fig.add_subplot(gs[1])
    ls_ax = fig.add_subplot(gs[2])
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # o-c Plot
    oc_ax.errorbar(epoch_no, ominusc, ominuscerr, marker='.',
                   elinewidth=0.8, color='dimgrey', linestyle='',
                   capsize=2, alpha=0.8, zorder=1)
    oc_ax.scatter(epoch_no, ominusc, marker='.', zorder=2,
                  c=epoch_no)
    oc_ax.axhline(0, color='black', linestyle='--', linewidth=1)
    oc_ax.plot(fit_x, fit_y, color='red', alpha=0.8)
    red = 'dof'
    from matplotlib.offsetbox import AnchoredText
    txt = AnchoredText(f'$\chi^2_{{{red}}} = {chi2_red:.2f}\, ({nsig:.2f}\sigma)$' +\
                        f'\n$\mu_{{error}}= {loss:.2f}$',
                        loc='upper right', frameon=False,
                        prop=dict(fontweight="bold"))
    oc_ax.add_artist(txt)
    oc_ax.set_ylim([np.nanmin(ominusc)*1.5, np.nanmax(ominusc)*2])
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Phase Plot
    phase_ax.errorbar(epoch_phase, ominusc, ominuscerr, marker='.',
                      elinewidth=0.8, color='dimgrey', linestyle='', capsize=2,
                      alpha=0.8, zorder=1)
    phase_ax.scatter(epoch_phase, ominusc, c=epoch_no,
                     marker='.',  zorder=2, alpha=0.5)
    phase_ax.axhline(0, color='black', linestyle='--', linewidth=1)
    phase_ax.plot(fit_x_phase[np.argsort(fit_x_phase)],
               fit_y[np.argsort(fit_x_phase)], color='red', alpha=0.8)
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # LS Plot
    ls_ax.plot(P/frequency, power, linestyle='-', linewidth=0.75, zorder=2,
               marker='', color='k')
    
    pos = P * best_P * 1.97 #len(epoch_no)*0.985
    for (level, i) in zip(false_alarm_levels, levels):
            ls_ax.axhline(level, color='r', linestyle='--', alpha=0.8)
            ls_ax.annotate(f'{str(int(i*100))}'+'$\%$',
                           (pos, level*1.018), color='k', ha='center')
    ls_ax.annotate(f'Peak: {P*best_P:.2f}\n{fap*100:.2f}$\%$',
                        (P*best_P, power.max()*1.03), color='k', weight='bold', ha='center')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Labels
    oc_ax.set_xlabel('Period (Days)')
    oc_ax.set_ylabel('O-C (Minutes)')
    phase_ax.set_xlabel('Phase')
    phase_ax.set_ylabel('O-C (Minutes)')

    ls_ax.set_xlabel('Period (Days)')
    ls_ax.set_ylabel('Power')

    oc_ax.tick_params('both', which='both', direction='in', bottom=True, left=True)
    phase_ax.tick_params('both', which='both', direction='in', bottom=True, left=True)
    ls_ax.tick_params('both', which='both', direction='in', bottom=True, left=True)
    
    ls_ax.set_xlim([0, P * best_P * 2])
    
    upper_y_fap = false_alarm_levels[2] * 1.5
    upper_y_pow = max(power) * 1.2
    upper_y = np.nanmax([upper_y_fap, upper_y_pow])
    ls_ax.set_ylim([-0.01, upper_y])
    ls_ax.set_xscale('linear')

    fig.tight_layout()
    if exoplanet==None:
        fig.savefig('O-C_fold.jpg', bbox_inches='tight')
    else:
        if longterm==True:
            fig.savefig(f"firefly/{exoplanet}/{exoplanet} o-c longterm.jpg",
                    bbox_inches='tight')
        else:
            fig.savefig(f"firefly/{exoplanet}/{exoplanet} o-c.jpg",
                    bbox_inches='tight')
    return chi2_red, nsig, loss, fap, best_P
    
  
def read_fitted_lc(exoplanet, transits):
    '''
    Reads in lc's and cleans the data.

    '''
    from transitfit.lightcurve import LightCurve
    epoch_no = transits
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Read in lc's
    df = pd.DataFrame(columns=(['Phase', 'Normalised flux',
                                'Flux uncertainty','Best fit curve']))
    for i in range(0,epoch_no):
        file = f'firefly/{exoplanet}/fitted_lightcurves/t0_f0_e{i}_detrended.csv'
        lc = pd.read_csv(file)[['Phase', 'Normalised flux',
                                'Flux uncertainty','Best fit curve']]
        df = df.append(lc, ignore_index=True)
    
    df = df.sort_values('Phase').reset_index(drop=True)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Zoom in on transits
    fit_combined = df['Best fit curve']
    transit_mask = np.ma.masked_values(fit_combined, 1.).mask
    transit_loc = np.where(transit_mask==False)
    window = int(len(transit_loc[0]) * 2.5)
    # Check if the transit is still in baseline of data, and there is a transit
    if window==0:
        start, stop = 0, len(fit_combined)
    try:
        if ( (transit_loc[0][0] + window > len(fit_combined)) 
            or (transit_loc[0][0] - window < 0) 
            or (transit_mask).all() ):
            start, stop = 0, len(fit_combined)
    except:
        pass
    else:
        start, stop = transit_loc[0][0] - window, transit_loc[0][-1] + window
        if start < 0:
            start, stop = 0, len(fit_combined)
    transit_mask[start:stop] = False
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Apply Mask
    df_transit_mask = df[~transit_mask]
    # Check if the data masked leaves a significant transit
    if len(df_transit_mask) < int(0.05*len(fit_combined)):
        df_transit_mask = df
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Sigma clipping
    from astropy.stats.sigma_clipping import sigma_clip
    from astropy.stats.funcs import mad_std
    
    fluxdiff = np.max(df['Best fit curve']) - np.min(df['Best fit curve'])
    if fluxdiff < 0.001:
        sigma = 4
    elif 0.001 < fluxdiff < 0.005:
        sigma = 5
    else:
        sigma = 6
    df_transit_mask['Residual'] = df_transit_mask['Normalised flux'] \
                                  - df_transit_mask['Best fit curve']
    clip = sigma_clip(df_transit_mask['Residual'], sigma=sigma, 
                      stdfunc=mad_std).mask
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Apply Mask
    df_clipped_mask = df_transit_mask[~clip]
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Assign numpy arrays
    time_all = df_clipped_mask['Phase'].values
    flux_all = df_clipped_mask['Normalised flux'].values
    flux_err_all = df_clipped_mask['Flux uncertainty'].values
    fit_xall = df_clipped_mask['Phase'].values
    fit_yall = df_clipped_mask['Best fit curve'].values
    return time_all, flux_all, flux_err_all, fit_xall, fit_yall

def density_scatter(exoplanet, transits, P, cadence):
    """
    Scatter plot colored by 2d histogram
    """
    import numpy as np
    import matplotlib.pyplot as plt
    from astropy.stats.funcs import mad_std
    from astropy.stats.sigma_clipping import sigma_clip
    from scipy.interpolate import interpn
    from scipy import stats
    time_all, flux_all, flux_err_all, fit_xall, fit_yall = read_fitted_lc(exoplanet, transits)
    x, y, yerr = time_all, flux_all, flux_err_all
    diff = flux_all - fit_yall
    mad = np.std(diff)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    from transitfit.lightcurve import LightCurve
    
    cadence_min = cadence / 60
    cad_bin = cadence_min / (P * 60 * 24)
    lc = LightCurve(x, y, yerr)
    obs_length = x.max() - x.min()
    n_bins = int((obs_length)/cad_bin)
    bins=[n_bins,n_bins]
    binned_phase, binned_flux, binned_err, binned_residuals = lc.bin(cad_bin, diff)
    madbin = np.std(binned_residuals)
    obs_depth = 1 - np.min(fit_yall)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    data, x_e, y_e = np.histogram2d(x, y, bins=bins, density=True)
    z = interpn( ( 0.5*(x_e[1:] + x_e[:-1]), 0.5*(y_e[1:]+y_e[:-1]) ),
                data, np.vstack([x,y]).T, method="splinef2d",
                bounds_error=False)

    #To be sure to plot all data
    z[np.where(np.isnan(z))] = 0.0

    # Sort the points by density, so that the densest points are plotted last
    idx = z.argsort()
    x, y, z, diff = x[idx], y[idx], z[idx], diff[idx]
    cadences = len(y)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Histogram residuals
    unbinned_counts, bins = np.histogram(diff, bins=30)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Binned Histogram residuals
    binned_counts, _ = np.histogram(binned_residuals, bins)
    weighted_binned_counts = binned_counts * unbinned_counts.max()/binned_counts.max()
    from matplotlib.offsetbox import AnchoredText
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # All errorbars
    sns.set_theme(style="whitegrid")
    sns.set_context('paper')
    sns.set_palette('deep')
    txt = AnchoredText(f'$\sigma_{{unbinned}}={mad:.6f}$\n$' +\
                       f'\sigma_{{binned}}={madbin:.6f}$',
                       frameon=False,loc='lower right',
                       prop=dict(fontweight="bold"))
    fig = plt.subplots(figsize=(12,8))
    gs = gridspec.GridSpec(2, 2, height_ratios=[3,1], width_ratios=[10,1])
    ax = plt.subplot(gs[0])
    cb_ax = plt.subplot(gs[1])
    res_ax = plt.subplot(gs[2])
    hist_ax = plt.subplot(gs[3], sharey=res_ax)
    
    plt.set_cmap('hot')
    # Residuals
    res_ax.errorbar(x, diff, yerr, color='dimgrey',
                  alpha=0.1, zorder=1, capsize=2, ls='none')
    res_ax.scatter(x, diff, s=5, c=z, edgecolor='none', zorder=2)
    
    res_ax.scatter(binned_phase, binned_residuals,
                   s=5, alpha=1, color='white', zorder=4, edgecolor='none')
    res_ax.axhline(y=0, linestyle='--', color='k', zorder=5, lw=0.75)
    
    plt.set_cmap('hot')
    # Light Curve
    ax.errorbar(x, y, yerr, color='dimgrey',
                  alpha=0.1, zorder=1, capsize=2, ls='none')
    ax.scatter(x, y, c=z, zorder=2, s=5, edgecolor='none')
    ax.add_artist(txt)
    
    ax.scatter(binned_phase, binned_flux, zorder=4, s=5,
                color='white', edgecolor='none', alpha = 0.9)
    ax.plot(fit_xall, fit_yall, marker='', color='k', zorder=5, lw=1)
    # Histogram
    rgba_color = colors.to_rgba('dimgrey')
    facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)
    hist_ax.hist(bins[:-1], bins, weights=unbinned_counts,
                 orientation='horizontal', color=facecolor,
                 edgecolor=rgba_color, histtype='stepfilled')
    hist_ax.axhline(y=0, linestyle='--', color='k', zorder=3, lw=0.75)
    
    rgba_color = colors.to_rgba('red')
    facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)
    hist_ax.hist(bins[:-1], bins, weights=weighted_binned_counts,
                 orientation='horizontal', color=facecolor,
                 edgecolor=rgba_color, histtype='stepfilled',alpha=0.5)
    # Labels
    ax.xaxis.set_ticklabels([])
    cb_ax.axis('off')
    position = cb_ax.inset_axes([0.1,0.1,0.2,0.8])
    hist_ax.tick_params('both', which='both', direction='in',
                                 labelleft=False, labelbottom=False,
                                 right=False, top=False)
    res_ax.set_xlabel('Phase')
    ax.set_ylabel('Normalised Flux')
    res_ax.set_ylabel('Residual')
    from matplotlib.cm import ScalarMappable
    m = ScalarMappable(cmap='hot')
    cbar = plt.colorbar(m, ax=cb_ax, shrink=0.3, cax=position)
    cbar.set_label('Normalised Flux Density')
    plt.subplots_adjust(hspace=.0)
    plt.subplots_adjust(wspace=.0)
    fig[0].savefig(f'firefly/{exoplanet}/{exoplanet} density.png',
                   bbox_inches='tight')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # No errorbar
    sns.set_theme(style="darkgrid")
    sns.set_context('paper')
    sns.set_palette('deep')
    txt = AnchoredText(f'$\sigma_{{unbinned}}={mad:.6f}$\n$' +\
                       f'\sigma_{{binned}}={madbin:.6f}$',
                       frameon=False,loc='lower right',
                       prop=dict(fontweight="bold"))
    fig = plt.subplots(figsize=(12,8))
    gs = gridspec.GridSpec(2, 2, height_ratios=[3,1], width_ratios=[10,1])
    ax = plt.subplot(gs[0])
    cb_ax = plt.subplot(gs[1])
    res_ax = plt.subplot(gs[2])
    hist_ax = plt.subplot(gs[3], sharey=res_ax)
    
    plt.set_cmap('hot')
    res_ax.scatter(x, diff, s=5, alpha=0.8, c=z, edgecolor='none', zorder=1)
    res_ax.scatter(binned_phase, binned_residuals,
                   s=5, alpha=0.8, edgecolor='none', color='white', zorder=2)
    res_ax.axhline(y=0, color='k', linestyle='--', zorder=3, lw=0.75)
    # Lightcurve
    plt.set_cmap('hot')
    ax.xaxis.set_ticklabels([])
    ax.scatter(x, y, c=z, zorder=2, s=5, edgecolor='none')
    ax.plot(fit_xall, fit_yall, marker='', color='k', zorder=4, lw=1)
    ax.scatter(binned_phase, binned_flux, zorder=3, s=5, color='white',
               edgecolor='none')
    ax.add_artist(txt)
    # Histogram
    rgba_color = colors.to_rgba('dimgrey')
    facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)
    hist_ax.hist(bins[:-1], bins, weights=unbinned_counts,
                 orientation='horizontal', color=facecolor,
                 edgecolor=rgba_color, histtype='stepfilled')
    hist_ax.axhline(y=0, linestyle='--', color='k', zorder=3, lw=0.75)
    
    rgba_color = colors.to_rgba('red')
    facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)
    hist_ax.hist(bins[:-1], bins, weights=weighted_binned_counts,
                 orientation='horizontal', color=facecolor,
                 edgecolor=rgba_color, histtype='stepfilled',alpha=0.5)
   # Labels
    ax.xaxis.set_ticklabels([])
    cb_ax.axis('off')
    position = cb_ax.inset_axes([0.1,0.1,0.2,0.8])
    hist_ax.tick_params('both', which='both', direction='in',
                                 labelleft=False, labelbottom=False,
                                 right=False, top=False)
    res_ax.set_xlabel('Phase')
    ax.set_ylabel('Normalised Flux')
    res_ax.set_ylabel('Residual')
    from matplotlib.cm import ScalarMappable
    m = ScalarMappable(cmap='hot')
    cbar = plt.colorbar(m, ax=cb_ax, shrink=0.3, cax=position)
    cbar.set_label('Normalised Flux Density')
    plt.subplots_adjust(hspace=.0)
    plt.subplots_adjust(wspace=.0)
    fig[0].savefig(f'firefly/{exoplanet}/{exoplanet} density noerr.png',
                   bbox_inches='tight')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # No resid or err
    sns.set_theme(style="darkgrid")
    sns.set_context('paper')
    sns.set_palette('deep')
    txt = AnchoredText(f'$\sigma_{{unbinned}}={mad:.6f}$\n$' +\
                       f'\sigma_{{binned}}={madbin:.6f}$',
                       frameon=False,loc='lower right',
                       prop=dict(fontweight="bold"))
    fig = plt.subplots(figsize=(12,6))
    gs = gridspec.GridSpec(1, 1)
    ax = plt.subplot(gs[0])
    
    plt.set_cmap('hot')
    ax.scatter(x, y, c=z, zorder=2, s=5, edgecolor='none')
    ax.scatter(binned_phase, binned_flux, zorder=3, s=5, color='white',
               edgecolor='none')
    ax.plot(fit_xall, fit_yall, marker='', color='k', zorder=3, lw=1)
    ax.add_artist(txt)
    plt.xlabel('Phase')
    ax.set_ylabel('Normalised Flux')
    fig[0].savefig(f'firefly/{exoplanet}/{exoplanet} density noresiderr.png',
                   bbox_inches='tight')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # No resid
    sns.set_theme(style="whitegrid")
    sns.set_context('paper')
    sns.set_palette('deep')
    txt = AnchoredText(f'$\sigma_{{unbinned}}={mad:.6f}$\n$' +\
                       f'\sigma_{{binned}}={madbin:.6f}$',
                       frameon=False,loc='lower right',
                       prop=dict(fontweight="bold"))
    fig = plt.subplots(figsize=(12,6))
    gs = gridspec.GridSpec(1, 1)
    ax = plt.subplot(gs[0])
    
    plt.set_cmap('hot')
    ax.errorbar(x, y, yerr, color='dimgrey',
                  alpha=0.1, zorder=1, capsize=2, ls='none')
    ax.scatter(x, y, c=z, zorder=2, s=5, edgecolor='none')
    ax.scatter(binned_phase, binned_flux, zorder=4, s=5, color='white',
               edgecolor='none')
    ax.plot(fit_xall, fit_yall, marker='', color='k', zorder=3, lw=1)
    # ax.errorbar(stat[1][:len(stat[0])], stat[0], stat_err[0]/6, color='yellow',
    #               alpha=0.7, capsize=2, ls='none', zorder=3)
    ax.add_artist(txt)
    plt.xlabel('Phase')
    ax.set_ylabel('Normalised Flux')
    fig[0].savefig(f'firefly/{exoplanet}/{exoplanet} density noresid.png',
                   bbox_inches='tight')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    return mad, madbin, obs_depth, cadences


def plot_coupled_comparison():
    here = os.path.dirname(os.path.abspath(__file__))
    here = ''
    coupled = pd.read_csv(f'{here}data/spear.csv')
    uncoupled = pd.read_csv(f'{here}data/spear_uncoupled.csv').set_index('pl_name').drop('WD 1856+534 b', axis=0)
    uncoupled = uncoupled.reset_index()
    
    pl_name_coupled = coupled['pl_name']
    pl_name_uncoupled = uncoupled['pl_name']
    
    missing_uncoup = np.setdiff1d(pl_name_uncoupled, pl_name_coupled).tolist()
    missing_coup = np.setdiff1d(pl_name_coupled, pl_name_uncoupled).tolist()
    uncoupled = pd.read_csv(f'{here}data/spear_uncoupled.csv').set_index('pl_name').drop(missing_uncoup, axis=0).drop('WD 1856+534 b', axis=0)
    uncoupled = uncoupled.reset_index()
    coupled = pd.read_csv(f'{here}data/spear.csv').set_index('pl_name').drop(missing_coup, axis=0)
    coupled = coupled.reset_index()
    
    
    pl_rad_coupled = coupled['pl_orbsmax']
    pl_rad_uncoupled = uncoupled['pl_orbsmax']
    
    resid = pl_rad_coupled - pl_rad_uncoupled
    
    fig = plt.subplots(figsize=(12,8))
    gs = gridspec.GridSpec(2, 2, height_ratios=[2,2], width_ratios=[10,1])
    ax = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[2], sharex=ax)
    
    
    ax.plot(pl_rad_coupled, pl_rad_uncoupled, marker='.', lw=0)
    ax.plot(pl_rad_uncoupled, color='r', marker='', lw=0.5)    
    ax.set_xlabel('Exoplanet')
    ax.set_ylabel(r'$r_p/r_*$')
    ax.legend(['Coupled', 'Uncoupled'])
    
    ax2.plot(resid, marker='.', lw=0)
    ax2.axhline(y=0, linestyle='--', color='k', zorder=3, lw=0.75)
    ax2.set_xlabel('Exoplanet')
    ax2.set_ylabel(r'$r_p/r_*$ coupled - uncoupled')
    plt.subplots_adjust(hspace=.0)
    plt.subplots_adjust(wspace=.0)
    
    
    return

def plot_ld_params():
    here = ''
    
    uncoupled_rank_sens = pd.read_csv(f'{here}data/spear_uncoupled.csv').sort_values(by='Sensitivity', ascending=False).set_index('pl_name').drop('WD 1856+534 b', axis=0)
    top10 = uncoupled_rank_sens[0:40].reset_index()['pl_name']
    
    coupled = pd.read_csv(f'{here}spear_coupled_ld_params.csv')
    uncoupled = pd.read_csv(f'{here}spear_uncoupled_ld_params.csv').set_index('pl_name').drop('WD 1856+534 b', axis=0)
    uncoupled = uncoupled.reset_index()
    
    pl_name_coupled = coupled['pl_name']
    pl_name_uncoupled = uncoupled['pl_name']
    
    missing_uncoup = np.setdiff1d(pl_name_uncoupled, pl_name_coupled).tolist()
    missing_coup = np.setdiff1d(pl_name_coupled, pl_name_uncoupled).tolist()
    uncoupled = pd.read_csv(f'{here}spear_uncoupled_ld_params.csv').set_index('pl_name').drop(missing_uncoup, axis=0).drop('WD 1856+534 b', axis=0)
    uncoupled = uncoupled.reset_index()
    coupled = pd.read_csv(f'{here}spear_coupled_ld_params.csv').set_index('pl_name').drop(missing_coup, axis=0)
    coupled = coupled.reset_index()
    coupled_filter = coupled['pl_name'].isin(top10)
    
    top10_coup = coupled[coupled_filter]
    top10_uncoup = uncoupled[coupled_filter]
    
    # LD PARAMS
    q0_coupled = coupled['q0']
    q0err_coupled = coupled['q0err']
    q1_coupled = coupled['q1']
    q1err_coupled = coupled['q1err']
    q0_uncoupled = uncoupled['q0']
    q0err_uncoupled = uncoupled['q0err']
    q1_uncoupled = uncoupled['q1']
    q1err_uncoupled = uncoupled['q1err']
    
    q0_coupled_top10 = top10_coup['q0']
    q0err_coupled_top10 = top10_coup['q0err']
    q1_coupled_top10 = top10_coup['q1']
    q1err_coupled_top10 = top10_coup['q1err']
    q0_uncoupled_top10 = top10_uncoup['q0']
    q0err_uncoupled_top10 = top10_uncoup['q0err']
    q1_uncoupled_top10 = top10_uncoup['q1']
    q1err_uncoupled_top10 = top10_uncoup['q1err']
    
    # PLOT
    fig = plt.subplots(figsize=(16,8))
    gs = gridspec.GridSpec(4, 2, height_ratios=[2,2,1,1])
    
    # q0
    q0 = plt.subplot(gs[0])
    q0.plot(q0_coupled, marker='.', lw=0)
    q0.plot(q0_uncoupled, color='r', marker='.', lw=0)  
    
    q0.set_xlabel('Exoplanet')
    q0.set_ylabel(r'$q_0$')
    q0.legend(['Coupled', 'Uncoupled'])
    
    # q1
    q1 = plt.subplot(gs[2], sharex=q0)
    q1.plot(q1_coupled, marker='.', lw=0)
    q1.plot(q1_uncoupled, color='r', marker='.', lw=0) 
    
    q1.set_xlabel('Exoplanet')
    q1.set_ylabel(r'$q_1$')
    #q1.legend(['Coupled', 'Uncoupled'])
    
    plt.subplots_adjust(hspace=.0)
    plt.subplots_adjust(wspace=.0)
    
    # resid
    residq0 = q0_coupled - q0_uncoupled
    residq1 = q1_coupled - q1_uncoupled
    resq0 = plt.subplot(gs[4], sharex=q0)
    resq0.plot(residq0, marker='.', lw=0, color='k')
    resq0.axhline(y=0, linestyle='--', color='k', zorder=3, lw=0.75)
    resq0.set_ylabel(r'$q_0$ c - u')
    
    resq1 = plt.subplot(gs[6], sharex=q0)
    resq1.plot(residq1, marker='.', lw=0, color='k')
    resq1.axhline(y=0, linestyle='--', color='k', zorder=3, lw=0.75)
    resq1.set_xlabel('Exoplanet')
    resq1.set_ylabel(r'$q_1$ c - u')
    
    # PLOT 2
    fig = plt.subplots(figsize=(12,8))
    gs = gridspec.GridSpec(4, 2, height_ratios=[2,2,1,1])
    
    # q0
    q0 = plt.subplot(gs[0])
    q0.scatter(q0_coupled, q0_uncoupled, marker='.', lw=0, alpha=0.5, edgecolor='none')
    q0.scatter(q0_coupled_top10, q0_uncoupled_top10, marker='.', lw=0, color='k')
    q0.set_xlabel(r'$q_0$ coupled')
    q0.set_ylabel(r'$q_0$ uncoupled')
    
    r = np.corrcoef(q0_coupled_top10, q0_uncoupled_top10)[0,1]
    from matplotlib.offsetbox import AnchoredText
    txt = AnchoredText(f'$R=${r:.2f}',
                       frameon=False,loc='lower right',
                       prop=dict(fontweight="bold"))
    q0.add_artist(txt)
    m, b = np.polyfit(q0_coupled_top10, q0_uncoupled_top10, 1)
    q0.plot(q0_coupled_top10, m*q0_coupled_top10+b, marker='', lw=0.5, color='k')
    # q1
    q1 = plt.subplot(gs[2])
    q1.scatter(q1_coupled, q1_uncoupled, marker='.', lw=0, alpha=0.5, edgecolor='none')
    q1.scatter(q1_coupled_top10, q1_uncoupled_top10, marker='.', lw=0, color='k')
    q1.set_xlabel(r'$q_1$ coupled')
    q1.set_ylabel(r'$q_1$ uncoupled')
    
    r = np.corrcoef(q1_coupled_top10, q1_uncoupled_top10)[0,1]
    txt = AnchoredText(f'$R=${r:.2f}',
                       frameon=False,loc='lower right',
                       prop=dict(fontweight="bold"))
    q1.add_artist(txt)
    m, b = np.polyfit(q1_coupled_top10, q1_uncoupled_top10, 1)
    q1.plot(q1_coupled_top10, m*q1_coupled_top10+b, marker='', lw=0.5, color='k')
    






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The backend for auto_retrieval.

@author: Steven Charles-Mindoza
"""

from ._archive import priors, _tic, _load_csv, _search
from ._plot import oc_fold, density_scatter

from rich import print
from transitfit import split_lightcurve_file, run_retrieval
from transitfit.plotting import quick_plot
from transitfit.lightcurve import LightCurve
from astroquery.mast import Observations as obs
from tqdm import tqdm
from datetime import datetime
from tabulate import tabulate
from astropy.config import set_temp_cache
from astropy.table import Table
from pandas import DataFrame, read_csv, Categorical
from shutil import rmtree, make_archive, copy
from natsort import natsorted
from math import ceil
import numpy as np
import sys
import os
import random



def _TESS_filter():
    here = os.path.dirname(os.path.abspath(__file__))
    os.makedirs('firefly/data', exist_ok=True)
    tess_filter_path = 'firefly/data/TESS_filter_path.csv'
    try:
        os.remove(tess_filter_path)
    except:
        pass
    tess_filter = f'{here}/data/Filters/TESS_filter.csv'
    cols = ['filter_idx', 'low_wl', 'high_wl']
    df = DataFrame(columns=cols)
    df = df.append([{'filter_idx': 0,
                     'low_wl': tess_filter}], ignore_index=True)
    df.to_csv(tess_filter_path, index=False, header=True)


def _alias(exoplanet):
    alias = read_csv('https://exoplanetarchive.ipac.caltech.edu/cgi-bin/' +\
        f'nstedAPI/nph-nstedAPI?table=aliastable&objname={exoplanet[:-2]}')
    check_wasp = alias['aliasdis'][1]
    if '1SWASP' in check_wasp:
        return check_wasp
    else:
        return None

def mast(exoplanet):
    _load_csv()
    highest, ratios = _search(exoplanet)
    exoplanet = highest[0]
    tic_id = _tic(exoplanet).replace('TIC ', '')
    print(f'\nSearching MAST for {exoplanet} (TIC {tic_id}).')
    search = obs.query_criteria(dataproduct_type=['timeseries'],
                                project='TESS',
                                target_name=tic_id
                                ).to_pandas()
    data = search[['obs_id', 'target_name', 't_exptime',
               'provenance_name', 'project']]
    data = data.rename(columns={"obs_id": "Product",
                                "target_name": "TIC ID",
                                "t_exptime": "Cadence",
                                "provenance_name": "HLSP",
                                "project": "Mission"})
    data['Product'] = \
                Categorical(data['Product'],
                ordered=True,
                categories=natsorted(data['Product'].unique()))
    data = data.sort_values('Product')
    print(f'\nQuery from MAST returned {len(search)} '
          f'data products for {exoplanet} (TIC {tic_id}).\n')
    return print(tabulate(data, tablefmt='psql', showindex=False, headers='keys'))


def discover():
    here = os.path.dirname(os.path.abspath(__file__))
    os.makedirs('firefly/discover', exist_ok=True)
    upper_ecliptic = f'{here}/data/Candidates/19_sector_candidates.csv'
    #upper_ecliptic = '19_sector_candidates.csv'
    tic_ids = read_csv(upper_ecliptic)['TIC ID'].tolist()
    for i, tic_id in enumerate(tic_ids):
        _discover_search(tic_id)
        
        
def _discover_search(tic_id):
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # MAST Search
    search = obs.query_criteria(dataproduct_type=['timeseries'],
                                project='TESS',
                                provenance_name='SPOC',
                                t_exptime=120,
                                target_name=tic_id
                                ).to_pandas()
    data = search[['obs_id', 'target_name', 'dataURL', 't_exptime',
                   'provenance_name', 'project']]
    data['dataURL'] = ['https://mast.stsci.edu/api/v0.1/Download/file/?uri=' +\
                         data['dataURL'][i] for i in range(len(search))]
    # data = data[data['t_exptime'].isin(cadence)]
    data = data[~data.dataURL.str.endswith('_dvt.fits')].reset_index(drop=True)
    provenance_name = data['provenance_name'].tolist()
    lc_links = data['dataURL'].tolist()
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Extract Time series
    os.makedirs('firefly/discover', exist_ok=True)
    csv_in_dir = []
    print(f'Fitting TIC {tic_id}')
    for j, fitsfile in enumerate(lc_links):
        with fits.open(fitsfile, cache=False) as TESS_fits:
            time = TESS_fits[1].data['TIME'] + 2457000
            flux = TESS_fits[1].data['PDCSAP_FLUX']
            flux_err = TESS_fits[1].data['PDCSAP_FLUX_ERR']
            
        write_dict = []
        for i in range(len(time)):
            write_dict.append({'Time': time[i], 'Flux': flux[i],
                               'Flux err': flux_err[i]})
        source = f'firefly/discover/{tic_id}'
        mast_name = data['obs_id'][j]
        os.makedirs(f'{source}', exist_ok=True)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Extract all light curves to a single csv file
        csv_name = f'{source}/{mast_name}/{mast_name}.csv'
        write_dict = {'Time': time, 'Flux': flux, 'Flux err': flux_err}
        df = DataFrame(write_dict)
        df.to_csv(csv_name, index=False, na_rep='nan')
        csv_in_dir.append(f'{os.getcwd()}/{csv_name}')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Plot the final output
    for k, csv in enumerate(csv_in_dir):
        lc_plot(csv)

def clip(df):
    from lightkurve import LightCurve
    from astropy.stats.funcs import mad_std
    time = df['TIME']
    flux = df['PDCSAP_FLUX']
    flux_err = df['PDCSAP_FLUX_ERR']
    lc = LightCurve(time, flux, flux_err)
    lc = lc.remove_outliers(sigma_upper=6, sigma_lower=50, stdfunc=mad_std)
    df_new = DataFrame()
    df_new['TIME'] = lc.time
    df_new['PDCSAP_FLUX'] = lc.flux
    df_new['PDCSAP_FLUX_ERR'] = lc.flux_err
    print(f'Sigma clipped {len(df)-len(df_new)} cadences.')
    return df_new

def _fits(exoplanet,
          exo_folder,
          cache,
          hlsp,
          cadence,
          bitmask
):
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # MAST Search
    tic_id = _tic(exoplanet).replace('TIC ', '')
    print(f'\nSearching MAST for {exoplanet} (TIC {tic_id}).')
    search = obs.query_criteria(dataproduct_type=['timeseries'],
                                project='TESS',
                                provenance_name=hlsp,
                                t_exptime=cadence,
                                target_name=tic_id
                                ).to_pandas()
    if len(search)==0:
        rmtree(exo_folder)
        print(f'Search result contains no data products for {exoplanet}.')
        sys.exit(f'Search result contains no data products for {exoplanet}.')
    data = search[['obs_id', 'target_name', 'dataURL', 't_exptime',
                   'provenance_name', 'project', 'sequence_number']]
    data['dataURL'] = ['https://mast.stsci.edu/api/v0.1/Download/file/?uri=' +\
                         data['dataURL'][i] for i in range(len(search))]
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Dataframe Checks
    data = data[data['t_exptime']==cadence]
    data = data[~data.dataURL.str.endswith('_dvt.fits')].reset_index(drop=True)
    provenance_name = data['provenance_name'].tolist()
    lc_links = data['dataURL'].tolist()
    print(f'\nQuery from MAST returned {len(lc_links)} '
          f'data products for {exoplanet} (TIC {tic_id}).\n')
    sector_list = natsorted(data['sequence_number'].values.tolist())
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Print search result
    show = data[['obs_id', 'target_name', 't_exptime',
                   'provenance_name', 'project']]
    show = show.rename(columns={"obs_id": "Product",
                                "target_name": "TIC ID",
                                "t_exptime": "Cadence",
                                "provenance_name": "HLSP",
                                "project": "Mission"})
    show['Product'] = \
            Categorical(show['Product'],
            ordered=True,
            categories=natsorted(show['Product'].unique()))
    show = show.sort_values('Product')
    print(tabulate(show, tablefmt='psql', showindex=False, headers='keys'))
    print()
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Extract Time series
    csv_in_dir = []
    os.makedirs('firefly/data/cache', exist_ok=True)
    for j, fitsfile in enumerate(lc_links):
        TESS_fits = None
        while TESS_fits is None:
            try:
                with set_temp_cache('firefly/data/cache'):
                    TESS_fits = Table.read(fitsfile, cache=cache)
            except:
                pass
            if TESS_fits is not None:
                pass
        source = f'{exo_folder}/mastDownload'
        mast_name = data['obs_id'][j]
        os.makedirs(f'{source}/{mast_name}', exist_ok=True)
        TESS_fits.write(f'{source}/{mast_name}/{mast_name}.fits',
                          overwrite=True)
        TESS_fits = TESS_fits.to_pandas()
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
        # Quality Masks
        AttitudeTweak = 1
        SafeMode = 2
        CoarsePoint = 4
        EarthPoint = 8
        Argabrightening = 16
        Desat = 32
        ApertureCosmic = 64
        ManualExclude = 128
        Discontinuity = 256
        ImpulsiveOutlier = 512
        CollateralCosmic = 1024
        #: The first stray light flag is set manually by MIT based on visual inspection.
        Straylight = 2048
        #: The second stray light flag is set automatically by Ames/SPOC based on background level thresholds.
        Straylight2 = 4096
        PlanetSearchExclude = 8192
        BadCalibrationExclude = 16384
        IsuffTargetsErrorCorr = 32768
        DEFAULT_BITMASK = [
            AttitudeTweak, SafeMode, CoarsePoint, EarthPoint, Desat, ManualExclude,
            ImpulsiveOutlier, Argabrightening, BadCalibrationExclude#, Straylight2
        ]
        HARD_BITMASK = [
            AttitudeTweak, SafeMode, CoarsePoint, EarthPoint, Desat, ManualExclude,
            ImpulsiveOutlier, Argabrightening, BadCalibrationExclude, Straylight2
        ]
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Extract all light curves to a single csv file
        TESS_fits['TIME'] = TESS_fits['TIME'] + 2457000
        TESS_fits['PDCSAP_FLUX'] = TESS_fits['PDCSAP_FLUX'] + 10000
        csv_name = f'{source}/{mast_name}/{mast_name}.csv'
        _df = TESS_fits[['TIME', 'PDCSAP_FLUX', 'PDCSAP_FLUX_ERR', 'QUALITY']]
        if bitmask=='default':
            df = _df[~_df['QUALITY'].isin(DEFAULT_BITMASK)].drop('QUALITY', axis=1)
            print(f'Removed {len(_df)-len(df)} bad cadences.')
        elif bitmask=='hard':
            df = _df[~_df['QUALITY'].isin(HARD_BITMASK)].drop('QUALITY', axis=1)
            print(f'Removed {len(_df)-len(df)} bad cadences.')
            # df = clip(df)
        else:
            df = _df.drop('QUALITY', axis=1)
        df = df.rename(columns = {'TIME':'Time',
                                  'PDCSAP_FLUX':'Flux',
                                  'PDCSAP_FLUX_ERR':'Flux err'})
        df.to_csv(csv_name, index=False, na_rep='nan')
        csv_in_dir.append(f'{os.getcwd()}/{csv_name}')
    csv_in_dir = natsorted(csv_in_dir)
    print('\nSplitting up the lightcurves into seperate epochs:\n')
    # csv_in_dir = []
    # for r, d, f in os.walk(source):
    #     for item in f:
    #         if '.csv' in item:
    #             csv_in_dir.append(os.path.join(r, item))
    return csv_in_dir, sector_list
    

def _retrieval(
        # Firefly Interface
        exoplanet,
        archive='eu',
        curve_sample=1,
        clean=False,
        cache=False,
        auto=True,
        # MAST Search
        hlsp=['SPOC'],
        cadence=120,
        bitmask='default',
        # TransitFit Variables
        walks=25,
        slices=5,
        cutoff=0.25,
        window=2.5,
        nlive=300,
        fit_ttv=False,
        detrending_list=[['nth order', 2]],
        dynesty_sample='auto',
        fitting_mode='auto',
        limb_darkening_model='quadratic',
        ld_fit_method='independent',
        max_batch_parameters=25,
        batch_overlap=2,
        dlogz=None,
        maxiter=None,
        maxcall=None,
        dynesty_bounding='multi',
        normalise=True,
        detrend=True,
        detrending_limits=[[-10,10]],
        # Plotting
        plot=True,
        marker_color='dimgray',
        line_color='black',
        bin_data=True,
        binned_color='red',
        nprocs=1,
):
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Filter Setup
    exo_folder = f'firefly/{exoplanet}'
    os.makedirs(exo_folder, exist_ok=True)
    _TESS_filter()
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Download Archive
    if fit_ttv==True:
        archive = 'spearnet'
    if archive=='nasa':
        host_T, host_z, host_r, host_logg, t0, P, t14, repack, ra, dec, dist, t_mag = \
            priors(exoplanet, archive=archive, save=True, user=False, auto=auto,
                    fit_ttv=fit_ttv)
    else:
        host_T, host_z, host_r, host_logg, t0, P, t14, repack = \
            priors(exoplanet, archive=archive, save=True, user=False, auto=auto,
                    fit_ttv=fit_ttv)
    is_nan = repack.isnull().values.any()
    if auto==False or is_nan==True:
        answer = ''
        while answer!='y':
            print('\nMissing values detected in priors located at:')
            print(f'{os.getcwd()}/{exo_folder}')
            answer = input('\nModify your priors file, then enter y to proceed. ')
            if answer=='q':
                sys.exit('Exiting..')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Split the Light curves
    split_curve_in_dir = []
    transits_per_sector = []
    csv_in_dir, sector_list = _fits(exoplanet, exo_folder=exo_folder, cache=cache, 
                                    hlsp=hlsp, cadence=cadence, bitmask=bitmask)
    for i, csvfile in enumerate(csv_in_dir):
        split_curves = split_lightcurve_file(csvfile, t0=t0, P=P, t14=t14,
                                             cutoff=cutoff, window=window)
        split_curves = [s + '.csv' for s in split_curves]
        
        split_curve_in_dir.append(split_curves)
        transits_per_sector.append(len(split_curves))
    split_curve_in_dir = [i for sub in split_curve_in_dir for i in sub]
    print(f'\nA total of {len(split_curve_in_dir)}', 
        'lightcurves' if len(split_curve_in_dir) > 1 else 'lightcurve',
        'were' if len(split_curve_in_dir) > 1 else 'was', 'generated.')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Enforce batches be the same size
    curves = ceil(curve_sample * len(split_curve_in_dir))
    if fit_ttv==True:
        equal_batches = [5+3*n for n in range(0,500)]
    else:
        equal_batches = [6+4*n for n in range(0,500)]
    if int(curves) > 5:
        if int(curves) not in equal_batches:
            import bisect
            index = bisect.bisect(equal_batches, int(curves))
            new_curves = equal_batches[index-1]
            print(f'\nEnforcing lightcurves to be in equal batch sizes. Discarded {curves-new_curves}.')
            curves = new_curves
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Sort the files into ascending order and take random sample
    split_curve_in_dir = random.sample(split_curve_in_dir, k=int(curves))
    split_curve_in_dir = natsorted(split_curve_in_dir)
    print(f'\nA total of {len(split_curve_in_dir)}', 
        'lightcurves' if len(split_curve_in_dir) > 1 else 'lightcurve',
        'will be used.')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Quick plot
    print(f'\nGenerating quick plots for {len(split_curve_in_dir)}',
        'lightcurves.' if len(split_curve_in_dir) > 1 else 'lightcurve.') 
    os.makedirs(f'{exo_folder}/quick_plots', exist_ok=True)
    for i, lc in enumerate(tqdm(split_curve_in_dir)):
        split = read_csv(lc)
        light_curve = LightCurve(split.Time, split.Flux, split.Flux_err)
        plot = quick_plot(light_curve, fname=f'split_curve_{i}', folder_path=f'{exo_folder}/quick_plots')
    print('\nSplit curve quick plots to check the transit window are located at:')
    print(f'{os.getcwd()}/{exo_folder}/quick_plots')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Set the Data Paths
    data_path = f'{exo_folder}/data_paths.csv'
    cols = ['Path', 'Telescope', 'Filter', 'Epochs', 'Detrending']
    df = DataFrame(columns=cols)
    for i, split_curve in enumerate(split_curve_in_dir):
        df = df.append([{'Path': split_curve}], ignore_index=True)
        df['Telescope'], df['Filter'], df['Detrending'] = 0, 0, 0
        df['Epochs'] = range(0, len(df))
    print('\nPassing all the data to TransitFit.\n')
    df.to_csv(data_path, index=False, header=True)
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Paths to data, priors, and filter info:
    data = data_path
    priors_csv = f'{exo_folder}/{exoplanet} Priors.csv'
    filters = 'firefly/data/TESS_filter_path.csv'
    os.makedirs('firefly/data/ldtk', exist_ok=True)
    ldtk_cache = 'firefly/data/ldtk'
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Output folders
    results_output_folder = f'{exo_folder}/output_parameters'
    fitted_lightcurve_folder = f'{exo_folder}/fitted_lightcurves'
    plot_folder = f'{exo_folder}/plots'
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Run the retrieval
    run_retrieval(
            data,
            priors_csv,
            filters,
            host_T=host_T,
            host_logg=host_logg,
            host_z=host_z,
            host_r=host_r,
            cadence=cadence/60,
            walks=walks,
            slices=slices,
            nlive=nlive,
            # fit_ttv=fit_ttv,
            allow_ttv=fit_ttv,
            detrending_list=detrending_list,
            dynesty_sample=dynesty_sample,
            fitting_mode=fitting_mode,
            limb_darkening_model=limb_darkening_model,
            ld_fit_method=ld_fit_method,
            ldtk_cache=ldtk_cache,
            max_batch_parameters=max_batch_parameters,
            batch_overlap=batch_overlap,
            dlogz=dlogz,
            maxiter=maxiter,
            maxcall=maxcall,
            dynesty_bounding=dynesty_bounding,
            normalise=normalise,
            detrend=detrend,
            detrending_limits=detrending_limits,
            results_output_folder=results_output_folder,
            final_lightcurve_folder=fitted_lightcurve_folder,
            plot_folder=plot_folder,
            plot=plot,
            marker_color=marker_color,
            line_color=line_color,
            bin_data=bin_data,
            binned_color=binned_color,
            n_procs=nprocs,
        )
    os.makedirs('firefly/plots', exist_ok=True)
    os.makedirs('firefly/plots/folded', exist_ok=True)
    os.makedirs('firefly/plots/density', exist_ok=True)
    os.makedirs('firefly/plots/densitynoerr', exist_ok=True)
    os.makedirs('firefly/plots/oc', exist_ok=True)
    try:
        copy(f'{exo_folder}/plots/folded_curves/with_errorbars/filter_0.png',
             f'firefly/plots/folded/{exoplanet}.png')
    except:
        print(f'Unable to overwrite {exoplanet}.jpg')
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Save Best Values
    master = read_csv(f'{exo_folder}/output_parameters/Complete_results.csv',
                      index_col='Parameter')
    master = master[['Best', 'Error']]
    P = master['Best']['P']
    q0 = master['Best']['q0']
    q0err = master['Error']['q0']
    q1 = master['Best']['q1']
    q1err = master['Error']['q1']
    u0 = master['Best']['u0']
    u0err = master['Error']['u0']
    u1 = master['Best']['u1']
    u1err = master['Error']['u1']
    try:
        Perr = float(master['Error'].filter(like = 'P', axis=0))
    except ValueError:
        Perr = float()
    if fit_ttv==True:
        t0 = master['Best']['t0'][0]
        t0err = float(master['Error']['t0'][0])
    else:
        t0 = master['Best']['t0']
        t0err = float(master['Error']['t0'])
    a = master['Best']['a/AU']
    aerr = float(master['Error']['a/AU'])
    ar = master['Best']['a/r*']
    arerr = float(master['Error']['a/r*'])
    rp = master['Best']['rp/r*']
    rperr = float(master['Error']['rp/r*'])
    inc = master['Best']['inc']
    incerr = float(master['Error']['inc'])
    ecc = master['Best']['ecc']
    try:
        eccerr = float(master['Error']['ecc'])
    except ValueError:
        eccerr = float()
    w = master['Best']['w']
    try:
        werr = float(master['Error']['w'])
    except ValueError:
        werr = float()
    try:
        mad, madbin, obs_depth, cadences = density_scatter(exoplanet=exoplanet,
                                transits=int(len(df)), P=P, cadence=cadence)
        copy(f'{exo_folder}/{exoplanet} density.png',
             f'firefly/plots/density/{exoplanet}.png')
        copy(f'{exo_folder}/{exoplanet} density noerr.png',
             f'firefly/plots/densitynoerr/{exoplanet}.png')
    except Exception as e:
        print(e)
    t_depth = rp**2
    sens = t_depth/mad
    sens_bin = t_depth/madbin
    obs_sens = obs_depth/mad
    obs_sens_bin = obs_depth/madbin
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Archive and sort
    print(
        'Variables used:\n\n'
        f'target={exoplanet}\n'
        f'archive={archive}\n'
        f'hlsp={hlsp}\n'
        f'curve_sample={str(curve_sample)}\n'
        f'clean={clean}\n'
        f'cache={cache}\n'
        f'cutoff={str(cutoff)}\n'
        f'window={str(window)}\n'
        f'nlive={str(nlive)}\n'
        f'fit_ttv={fit_ttv}\n'
        f'detrending_list={str(detrending_list)}\n'
        f'dynesty_sample={dynesty_sample}\n'
        f'fitting_mode={fitting_mode}\n'
        f'limb_darkening_model={limb_darkening_model}\n'
        f'ld_fit_method={ld_fit_method}\n'
        f'max_batch_parameters={str(max_batch_parameters)}\n'
        f'batch_overlap={str(batch_overlap)}\n'
        f'dlogz={str(dlogz)}\n'
        f'maxiter={str(maxiter)}\n'
        f'maxcall={str(maxcall)}\n'
        f'dynesty_bounding={dynesty_bounding}\n'
        f'normalise={normalise}\n'
        f'detrend={detrend}',
        file=open(exo_folder+'/variables.txt', 'w')
    )
    now = datetime.now().strftime("%d-%b-%Y %H:%M:%S")
    sci_prod = ' '.join(hlsp)
    archive_name = f"{exoplanet} {archive.upper()} {sci_prod} {now}"
    if clean==True:
        try:
            rmtree(f'{exo_folder}/output_parameters/quicksaves')
            rmtree(f'{exo_folder}/output_parameters/filter_0_parameters/quicksaves')
        except Exception:
            pass
    if fit_ttv==True:
        try:
            here = os.path.dirname(os.path.abspath(__file__))
            spearnet_csv = f'{here}/data/spear.csv'
            spearnet = read_csv(spearnet_csv).set_index('pl_name')
            s = spearnet.loc[[exoplanet]]
            t0ttv = s['pl_tranmid'][0]
            t0errttv = s['pl_tranmiderr1'][0]
            file = f'firefly/{exoplanet}/output_parameters/Complete_results.csv'
            chi2_redlong, nsiglong, losslong, faplong, periodlong = oc_fold(t0ttv, t0errttv,
                                                        file=file, P=P,
                                                        exoplanet=exoplanet)
            try:
                chi2_red, nsig, loss, fap, period = oc_fold(t0ttv, t0errttv,
                                                        file=file, P=P,
                                                        exoplanet=exoplanet,
                                                        longterm=False)
            except:
                chi2_red, nsig, loss, fap, period = float(), float(), float(), float()
                pass
            copy(f'{exo_folder}/{exoplanet} o-c longterm.jpg',
                 f'firefly/plots/oc/{exoplanet} long.jpg')
            copy(f'{exo_folder}/{exoplanet} o-c.jpg',
                 f'firefly/plots/oc/{exoplanet}.jpg')
            data = {'pl_name':exoplanet, 'red_chi2':chi2_red, 'sigma':nsig,
                    'mean_avg_err':loss,
                    'fap':fap, 'o-c_period':period,
                    'faplong':faplong, 'o-c_periodlong':periodlong,
                    'pl_orbper':P, 'pl_orbpererr1':Perr,
                    'pl_tranmid':t0, 'pl_tranmiderr1':t0err,
                    'pl_orbsmax':a, 'pl_orbsmaxerr1':aerr, 'pl_radj':rp,
                    'pl_radjerr1':rperr, 'pl_orbincl':inc,
                    'pl_orbinclerr1':incerr, 'pl_orbeccen':ecc, 'pl_orbeccenerr1':eccerr,
                    'pl_orblper':w, 'pl_orblpererr1':werr, 'q0':q0, 'q0err':q0err,
                    'q1':q1, 'q1err':q1err, 'u0':u0, 'u0err':u0err,
                    'u1':u1, 'u1err':u1err,
                    'Transits':int(len(df)),
                    'Date':now, 'Archive':archive.upper(), 'Unbinned Sigma':mad,
                    'Binned Sigma':madbin,
                'Transit Depth':t_depth, 'Sensitivity':sens,
                'Binned Sensitivity':sens_bin,
                'Observation Depth':obs_depth, 'Observation Sensitivity':obs_sens,
                'Binned Observation Sensitivity':obs_sens_bin, 'Total Cadences':cadences
            }
            df = DataFrame(data, index=[0])
            summary_master = 'firefly/data/spear_ttv.csv'
            if not os.path.exists(summary_master):
                df.to_csv(summary_master, index=False)
            else:
                add = read_csv(summary_master)
                add = add.append(df)
                add['pl_name'] = \
                    Categorical(add['pl_name'],
                    ordered=True,
                    categories=natsorted(add['pl_name'].unique()))
                add = add.sort_values('pl_name')
                add .to_csv(summary_master, index=False)
        except Exception as e:
            print(e, file=open(exo_folder+'/ttv_traceback.txt', 'w'))
        os.makedirs('firefly/ttv', exist_ok=True)
        os.makedirs(f'firefly/ttv/{fitting_mode}', exist_ok=True)
        archive_folder = f'firefly/ttv/{fitting_mode}/{archive_name}'
        make_archive(archive_folder, format='zip',
                     root_dir=f'{os.getcwd()}/firefly/',
                     base_dir=f'{exoplanet}')
    else:
        data = {'pl_name':exoplanet, 'pl_orbper':P, 'pl_orbpererr1':Perr,
                'pl_tranmid':t0, 'pl_tranmiderr1':t0err,
                'pl_orbsmax':a, 'pl_orbsmaxerr1':aerr, 'pl_radj':rp,
                'pl_radjerr1':rperr, 'pl_orbincl':inc,
                'pl_orbinclerr1':incerr, 'pl_orbeccen':ecc, 'pl_orbeccenerr1':eccerr,
                'pl_orblper':w, 'pl_orblpererr1':werr, 'q0':q0, 'q0err':q0err,
                'q1':q1, 'q1err':q1err, 'u0':u0, 'u0err':u0err,
                    'u1':u1, 'u1err':u1err,
                'Transits':int(len(df)),
                'Date':now, 'Archive':archive.upper(), 'Unbinned Sigma':mad,
                'Binned Sigma':madbin,
                'Transit Depth':t_depth, 'Sensitivity':sens,
                'Binned Sensitivity':sens_bin,
                'Observation Depth':obs_depth, 'Observation Sensitivity':obs_sens,
                'Binned Observation Sensitivity':obs_sens_bin, 'Total Cadences':cadences
        }
        df = DataFrame(data, index=[0])
        summary_master = 'firefly/data/spear.csv'
        if not os.path.exists(summary_master):
            df.to_csv(summary_master, index=False)
        else:
            add = read_csv(summary_master)
            add = add.append(df)
            add['pl_name'] = \
                Categorical(add['pl_name'],
                ordered=True,
                categories=natsorted(add['pl_name'].unique()))
            add = add.sort_values('pl_name')
            add .to_csv(summary_master, index=False)
        os.makedirs(f'firefly/{fitting_mode}', exist_ok=True)
        archive_folder = f'firefly/{fitting_mode}/{archive_name}'
        make_archive(archive_folder, format='zip',
                     root_dir=f'{os.getcwd()}/firefly/',
                     base_dir=f'{exoplanet}')
    rmtree(f'{exo_folder}')
    return archive_name



def spearnet_archive_ld_params(source='13_April_2021_folded_uncoupled'):
    import zipfile
    import pandas as pd
    # Find Files
    zip_in_dir = []
    for r, d, f in os.walk(source):
        for item in f:
            if '.zip' in item:
                zip_in_dir.append(os.path.join(r, item))
    
    # Open Files
    for i, zf in enumerate(zip_in_dir):
        with zipfile.ZipFile(zf) as zip:
            exoplanet = zip.namelist()[0].replace('/', '')
            with zip.open(f'{exoplanet}/output_parameters/Complete_results.csv') as myZip:
                df = pd.read_csv(myZip, index_col='Parameter')
                master = df[['Best', 'Error']]
                q0 = master['Best']['q0']
                q0err = master['Error']['q0']
                q1 = master['Best']['q1']
                q1err = master['Error']['q1']
                u0 = master['Best']['u0']
                u0err = master['Error']['u0']
                u1 = master['Best']['u1']
                u1err = master['Error']['u1']
                
                data = {'pl_name':exoplanet, 'q0':q0, 'q0err':q0err,
                        'q1':q1, 'q1err':q1err, 'u0':u0, 'u0err':u0err,
                        'u1':u1, 'u1err':u1err}
                df = DataFrame(data, index=[0])
                summary_master = 'spear.csv'
                if not os.path.exists(summary_master):
                    df.to_csv(summary_master, index=False)
                else:
                    add = read_csv(summary_master)
                    add = add.append(df)
                    add['pl_name'] = \
                        Categorical(add['pl_name'],
                        ordered=True,
                        categories=natsorted(add['pl_name'].unique()))
                    add = add.sort_values('pl_name')
                    add .to_csv(summary_master, index=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
                                        
                     (                   (     (        
                     )\ )  (   (      (  )\ )  )\ (     
                    (()/(  )\  )(    ))\(()/( ((_))\ )  
                     /(_))((_)(()\  /((_)/(_)) _ (()/(  
                    (_) _| (_) ((_)(_)) (_) _|| | )(_)) 
                     |  _| | || '_|/ -_) |  _|| || || | 
                     |_|   |_||_|  \___| |_|  |_| \_, | 
                                                  |__/  
                                  
                  A data retriever for use with TransitFit.

    Functions
    ---------
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    firefly()
    
        Automated version of retrieval. For a single target the procedure is:
         
         >>> from firefly import firefly
             target = ('WASP-43 b',)
             firefly(target)
         
        For a list of targets:
             
             >>> from firefly import firefly
                 targets = ('WASP-43 b', 'WASP-18 b')
                 firefly(targets)
         
        - Targets passed are corrected for basic user input. 'wasp43b' is
          interpreted as 'WASP-43 b'. List must be of the form given in the 
          example below.
        - Initial checks for targets from the exoplanet archive are then taken 
          to ascertain whether the prior data extracted has entries in all 
          columns. If there are missing entries for a given target in the list, 
          the user will be asked whether to proceed.
        - Iteratively takes the targets given and employs TransitFit across 
          each TESS sector for every exoplanet in the list given.
        - All available split curves are fitted with TransitFit, then the 
          results are then zipped up and time stamped. Optionally sends an 
          email upon an error or full completion of a target.
        - Email is also disabled by default. If enabled, status updates on 
          completion and exceptions with the full traceback are sent. 
          
        Background tasks for feeding data to TransitFit include:
        
        - Set the filter path to the TESS Filter.
        - Download EU/NASA exoplanet archive data every 10 days 
          (checks the file age).
        - Download MAST lightcurve data for target TESS Sectors.
        - Split the lightcurves into seperate transits or epochs.
        - Create the data paths to each seperate epoch.
        - Run TransitFit.
        - Delete all downloaded data and inputs.
        - Zip and timestamp the output.
            
        Input is capable of handling :
        
            >>> ('wasp43b', 'WASp18b', 'wasP91--b') etc
            
        Forces corrections based on classifier: 
        
            >>> 'WASP', 'LTT', 'GJ' etc
            
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    
"""

name = 'firefly'
__version__ = '0.8.2'

from .auto_retrieval import firefly
from ._archive import priors, tess, tess_candidate, gen_tess
from ._plot import plot_epoch, oc, oc_fold
from ._utils import mast, discover



#!/usr/bin/env python3

import os
import codecs
import pathlib
import setuptools

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()
        
def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")
    
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup( name='firefly-tess',
                  version=get_version('firefly/__init__.py'),
                  description='For use with TransitFit',
                  author='Stephen Charles',
                  license='MIT',
                  long_description=long_description,
                  long_description_content_type='text/markdown',
                  packages=setuptools.find_packages(),
                  download_url = 'https://github.com/sourestdeeds/firefly/archive/refs/tags/v2.0.0.tar.gz',
                  classifiers=[
                    "Programming Language :: Python :: 3",
                    "License :: OSI Approved :: MIT License"
                  ],
                  install_requires=['numpy',
                                    'pandas', 
                                    'tabulate',
                                    'fuzzywuzzy',
                                    'python-Levenshtein',
                                    'natsort',
                                    'seaborn',
                                    'astroquery',
                                    'pyastronomy',
                                    'astropy',
                                    'lightkurve',
                                    'transitfit',
                                    'rich',
                                    'tqdm',
                                    ],
                  python_requires='>=3.6',
                  include_package_data=True )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Automated version of retrieval.

@author: Steven Charles-Mindoza
"""

from ._utils import _retrieval
from ._archive import _search, _load_csv

from rich import print
from datetime import datetime
from shutil import rmtree, make_archive
from traceback import format_exc
import sys
import os



def _auto_input_check(targets, curve_sample):
    if not (0 < curve_sample <= 1):
        sys.exit('The curve sample must be in the range 0 < curve_sample <= 1.')
    _load_csv()
    targets = ''.join(targets)
    highest, ratios = _search(targets)
    exoplanet = highest[0]
    print(f'\nTarget search chose {exoplanet}.')
    return exoplanet


def firefly(
        # Firefly Interface
        targets,
        archive='nasa',
        curve_sample=1,
        clean=True,
        cache=True,
        auto=True,
        # MAST Search
        hlsp=['SPOC'],
        cadence=120,
        bitmask='default',
        # TransitFit Variables
        walks=100,
        slices=10,
        cutoff=0.25,
        window=12,
        nlive=2000,
        fit_ttv=False,
        detrending_list=[['nth order', 2]],
        dynesty_sample='rslice',
        fitting_mode='folded',
        limb_darkening_model='quadratic',
        ld_fit_method='coupled',
        max_batch_parameters=25,
        batch_overlap=2,
        dlogz=None,
        maxiter=None,
        maxcall=None,
        dynesty_bounding='multi',
        normalise=True,
        detrend=True,
        detrending_limits=[[-1000,1000]],
        # Plotting
        plot=True,
        marker_color='dimgray',
        line_color='black',
        bin_data=True,
        binned_color='red',
        nprocs=1,
):
    '''
    Automated version of retrieval. For a single target the procedure is:
         
         >>> from firefly import firefly
             target = ('WASP-43 b',)
             firefly(target)
         
    For a list of targets:
         
         >>> from firefly import firefly
             targets = ('WASP-43 b', 'WASP-18 b')
             firefly(targets)
     
    - Targets passed are corrected for basic user input; 'wasp43b' is
      interpreted as 'WASP-43 b'. List must be of the form given in the
      example below.
    - Initial checks for targets from the exoplanet archive are then taken
      to ascertain
      whether the prior data extracted has entries in all columns. If there
      are missing entries for a given target in the list, the user will be
      asked whether to proceed.
    - Iteratively takes the targets given and employs TransitFit across each
      TESS sector for every exoplanet in the list given.
    - All available split curves are fitted with TransitFit, then the results
      are then zipped up and time stamped. Optionally sends an email upon an
      error or full completion of a target.
    - Email is also disabled by default. If enabled, status updates on
      completion and exceptions with the full traceback are sent.
      
    Background tasks for feeding data to TransitFit include:
    
    - Set the filter path to the TESS Filter.
    - Download EU/NASA exoplanet archive data every 10 days
      (checks the file age).
    - Download MAST lightcurve data for target TESS Sectors.
    - Split the lightcurves into seperate transits or epochs.
    - Create the data paths to each seperate epoch.
    - Run TransitFit.
    - Delete all downloaded data and inputs.
    - Zip and timestamp the output.
        
    Input is capable of handling :
    
        >>> ('wasp43b', 'WASp18b', 'wasP91--b') etc
        
    Forces corrections based on classifier:
    
        >>> 'WASP', 'LTT', 'GJ' etc
    
    Parameters
    ----------
    targets : str, list
        A list of exoplanet targets.
        Input is a list tuple of strings:
            
        >>> ('WASP-43 b', 'WASP-18 b', 'WASP-91 b')
    curve_sample : int {0 < curve_sample <= 1}, optional
        The fraction of curves generated to fit against. For example, setting
        
        >>> curve_sample = 0.5
        
        will fit half the curves extracted. The formula for this works as:
            
        >>> total_curves =
            curves_extracted * curve_sample
        
        Always returns an int. For example:
            
        >>> curve_sample = 0.001
        
        will fit using only 1 lightcurve from each sector.
        The default is 1 to fit all lightcurves across all sectors.
    archive : str, {'eu', 'nasa', 'org', 'all'}
        The archive to generate priors from. All takes the IQR of all
        archives (including OEC) and then the mean.
        The default is 'eu'.
    clean : bool, optional
        If True will delete all downloaded files and zip outputs only.
        The default is False.
    cache : bool, optional
        If True will cache lightcurve fits files in root/.astropy.
        The default is False.
    auto : bool, optional
        If False will allow the user to modify the generated priors file
        before proceeding.
        The default is True.
    hlsp : str list, ['SPOC', 'TESS-SPOC', 'TASOC']
        SPOC is the primary TESS mission, and the rest are HLSP.
        The default is ['SPOC'].
    cadence : int list, [20, 120, 600, 1800]
        The exposure time. Options are 20, 120, 600 or 1800.
        The default is 120.
    bitmask : str, 'default' or 'hard'
        Filter out bad cadences.
        The default is 'default'.
        None will override the bitmask and use all cadences.
    cutoff : float, optional
        If there are no data within
        
        >>> t14 * cutoff of t0,
        
        a period will be
        discarded. Default is 0.25
    window : float, optional
        Data outside of the range
        
        >>> [t0 ± (0.5 * t14) * window]
        
        will be discarded.
    nlive : int, optional
        The number of live points to use in the nested sampling retrieval.
        Default is 1000.
    detrending_list : array_like, shape (n_detrending_models, 2)
        A list of different detrending models. Each entry should consist
        of a method and a second parameter dependent on the method.
        Accepted methods are:
            
        - ['nth order', order]
        - ['custom', function, [global fit indices, filter fit indices, epoch fit indices]]
        - ['off', ]
        Function here is a custom detrending function. TransitFit assumes
        that the first argument to this function is times and that all
        other arguments are single-valued - TransitFit cannot fit
        list/array variables. If 'off' is used, no detrending will be
        applied to the `LightCurve`s using this model.
        If a custom function is used, and some inputs to the function
        should not be fitted individually for each light curve, but should
        instead be shared either globally, within a given filter, or within
        a given epoch, the indices of where these fall within the arguments
        of the detrending function should be given as a list. If there are
        no indices to be given, then use an empty list: []
        e.g. if the detrending function is given by
            
        >>> foo(times, a, b, c):
                # do something
           
        and a should be fitted globally, then the entry in the method_list
        would be
        
        - ['custom', foo, [1], [], []].
    dynesty_sample : str, optional
        Method used to sample uniformly within the likelihood constraint,
        conditioned on the provided bounds. Unique methods available are:
            
        - uniform sampling within the bounds('unif')
        - random walks with fixed proposals ('rwalk')
        - random walks with variable ("staggering") proposals ('rstagger')
        - multivariate slice sampling along preferred orientations ('slice')
        - "random" slice sampling along all orientations ('rslice')
        - "Hamiltonian" slices along random trajectories ('hslice')
        and any callable function which follows
        the pattern of the sample methods defined in dynesty.sampling.
        'auto' selects the sampling method based on the dimensionality of
        the problem (from ndim).
        - When ndim < 10, this defaults to 'unif'.
        - When 10 <= ndim <= 20, this defaults to 'rwalk'.
        - When ndim > 20, this defaults to 'hslice' if a gradient is provided
          and 'slice' otherwise.
        - 'rstagger' and 'rslice' are provided as alternatives for
          'rwalk' and 'slice', respectively. Default is 'auto'.
    fitting_mode : {`'auto'`, `'all'`, `'folded'`, `'batched'`}, optional
        The approach TransitFit takes towards limiting the number of parameters
        being simultaneously fitted. The available modes are:
            
        - `'auto'` : Will calculate the number of parameters required to fit
          all the data simulataneously. If this is less than max_parameters,
          will set to `'all'` mode, else will set to `'folded'` if at least one
          filter has at least 3 epochs in it. Otherwise will set to `'batched'`
        - `'all'` : Fits all parameters simultaneously, with no folding or
          batching of curves. Should be used with caution when fitting very
          large (~< 30) numbers of parameters.
        - `'folded'` : Useful for fitting curves with multiple epochs for each
          filter. TransitFit will fit each filter separately and produce a
          period-folded light curve for each filter, before fitting these
          simultaneously, using the `'batched'` approach if required.
        - `'2_stage'` : Fits in 2 stages, first detrending the light curves and
          then fitting the detrended curves simultaneously, using the
          `'batched'` approach if required.
        - `'batched'` : Useful for large numbers of light curves with
          relatively few shared filters, so `'folded'` loses large amounts of
          multi-epoch information. This mode splits the filters into sets of
          overlapping batches, runs each batch and uses the weighted means of
          each batch to produce a final result.
        Default is `'auto'`.
    fit_ttv : boolean, optional
        DESCRIPTION. The default is False.
    limb_darkening_model : str, optional
        The limb darkening model to use. Allowed models are
            
        - 'linear'
        - 'quadratic'
        - 'squareroot'
        - 'power2'
        - 'nonlinear'
        With the exception of the non-linear model, all models are constrained
        by the method in Kipping (2013), which can be found at
        https://arxiv.org/abs/1308.0009. Use `ldc_low_lim` and `ldc_high_lim`
        to control the behaviour of unconstrained coefficients.
        Default is 'quadratic'.
    ld_fit_method : {`'coupled'`, `'single'`, `'independent'`, `'off'`}, optional
        Determines the mode of fitting of limb darkening parameters. The
        available modes are:
            
        - `'coupled'` : all limb darkening parameters are fitted
          independently, but are coupled to a wavelength dependent
          model based on the host parameters through `ldkt`
        - `'single'` : LD parameters are still tied to a model, but
          only the first filter is actively fitted. The remaining
          filters are estimated based off the ratios given by ldtk for
          a host with the given parameters. This mode is useful for a
          large number of filters, as `'coupled'` or `'independent'`
          fitting will lead to much higher computation times.
        - `'independent'` : Each LD coefficient is fitted separately for
          each filter, with no coupling to the ldtk models.
        - `'off'` : Will use the fixed value provided in the input file
        Default is `'independent'`
    max_batch_parameters : int, optional
        The maximum number of parameters to use in a single retrieval.
        Default is 25.
    batch_overlap : int, optional
        The number of epochs to overlap in each batch. This will be adhered
        to where possible. Default is 2.
    dlogz : float, optional
        Retrieval iteration will stop when the estimated contribution of
        the remaining prior volume to the total evidence falls below this
        threshold. Explicitly, the stopping criterion is
        
        >>> ln(z + z_est) - ln(z) < dlogz,
        
        where z is the current evidence
        from all saved samples and z_est is the estimated contribution from
        the remaining volume. The default is
        
        >>> 1e-3 * (nlive - 1) + 0.01.
    maxiter : int or `None`, optional
        The maximum number of iterations to run. If `None`, will
        continue until stopping criterion is reached. Default is `None`.
    maxcall : int or `None`, optional
        The maximum number of likelihood calls in retrieval. If None, will
        continue until stopping criterion is reached. Default is `None`.
    dynesty_bounding : {`'none'`, `'single'`, `'multi'`, `'balls'`, 'cubes'`}, optional
        The decomposition to use in sampling. Default is 'multi'
    normalise : bool, optional
        If True, will assume that the light curves have not been normalised and
        will fit normalisation constants within the retrieval. The range to
        fit normalisation constants c_n are automatically detected using
        
        >>> 1/f_min <= c_n <= 1/f_max
        
        as the default range, where f_min and f_max are the minimum and maximum
        flux values for a given light curve. Default is True.
    detrend : bool, optional
        If True, will initialise detrending fitting. Default is True.

    Returns
    -------
    A whole lot of data to science!
    Zipped files are found in:
    
    >>> firefly/WASP-43 b timestamp.gz.tar

    '''
    exoplanet = _auto_input_check(targets, curve_sample)
    try:
        archive_name = \
        _retrieval(
            # Firefly Interface
            exoplanet,
            archive=archive,
            curve_sample=curve_sample,
            clean=clean,
            cache=cache,
            auto=auto,
            # MAST Search
            hlsp=hlsp,
            cadence=cadence,
            bitmask=bitmask,
            # TransitFit Variables
            walks=walks,
            slices=slices,
            cutoff=cutoff,
            window=window,
            nlive=nlive,
            fit_ttv=fit_ttv,
            detrending_list=detrending_list,
            dynesty_sample=dynesty_sample,
            fitting_mode=fitting_mode,
            limb_darkening_model=limb_darkening_model,
            ld_fit_method=ld_fit_method,
            max_batch_parameters=max_batch_parameters,
            batch_overlap=batch_overlap,
            dlogz=dlogz,
            maxiter=maxiter,
            maxcall=maxcall,
            dynesty_bounding=dynesty_bounding,
            normalise=normalise,
            detrend=detrend,
            detrending_limits=detrending_limits,
            # Plotting
            plot=plot,
            marker_color=marker_color,
            line_color=line_color,
            bin_data=bin_data,
            binned_color=binned_color,
            nprocs=nprocs,
        )
        success = f'{os.getcwd()}/firefly/{archive_name}.zip'
        print(f'\nData location: {success}\n'
                'A new target has been fully retrieved across ' +
                'all available TESS Sectors.')
    except KeyboardInterrupt:
        exo_folder = f'firefly/{exoplanet}'
        now = datetime.now().strftime("%d-%b-%Y %H:%M:%S")
        keyboard = f'firefly/KeyboardInterrupt/{exoplanet} ' +\
                   f'{now} KeyboardInterrupt'
        make_archive(keyboard, format='zip',
                 root_dir=f'{os.getcwd()}/firefly/',
                 base_dir=f'{exoplanet}')
        rmtree(exo_folder)
        raise
        sys.exit()
    except BaseException as e:
        exo_folder = f'firefly/{exoplanet}'
        now = datetime.now().strftime("%d-%b-%Y %H:%M:%S")
        exception = f'firefly/Exception/{exoplanet} ' +\
                    f'{now} Exception'
        trace_back = format_exc()
        print(trace_back, file=open(exo_folder+'/traceback.txt', 'w'))
        print(
            'Variables used:\n\n'
            f'target={exoplanet}\n'
            f'archive={str(archive)}\n'
            f'curve_sample={str(curve_sample)}\n'
            f'clean={clean}\n'
            f'cache={cache}\n'
            f'auto={auto}\n'
            f'hlsp={str(hlsp)}\n'
            f'cadence={str(cadence)}\n'
            f'walks={str(walks)}\n'
            f'slices={str(slices)}\n'
            f'cutoff={str(cutoff)}\n'
            f'window={str(window)}\n'
            f'nlive={str(nlive)}\n'
            f'fit_ttv={fit_ttv}\n'
            f'detrending_list={str(detrending_list)}\n'
            f'dynesty_sample={dynesty_sample}\n'
            f'fitting_mode={fitting_mode}\n'
            f'limb_darkening_model={limb_darkening_model}\n'
            f'ld_fit_method={ld_fit_method}\n'
            f'max_batch_parameters={str(max_batch_parameters)}\n'
            f'batch_overlap={str(batch_overlap)}\n'
            f'dlogz={str(dlogz)}\n'
            f'maxiter={str(maxiter)}\n'
            f'maxcall={str(maxcall)}\n'
            f'dynesty_bounding={dynesty_bounding}\n'
            f'normalise={normalise}\n'
            f'detrend={detrend}',
            file=open(exo_folder+'/variables.txt', 'w')
        )
        make_archive(exception, format='zip',
                 root_dir=f'{os.getcwd()}/firefly/',
                 base_dir=f'{exoplanet}')
        rmtree(exo_folder)
        print(trace_back)
